import horus_gis as HG
import numpy as NP
import math

try:
    import shapely as SP
    from shapely import geometry
    import pyproj as PP
except ModuleNotFoundError as e:
    print(f"Install module '{e.name}' to use this module.")
    raise e
    
class Geometry_proj:

    class Projection:
        def __init__(self, model=None,is_enu=None,geometry=None,geod=None) -> None:
            self.model:HG.EnuModel = model
            self.is_enu:bool = is_enu
            self.geometry = geometry
            self.geod: PP.Geod = geod

    def create_projection(self,geom):                
        
        proj = Geometry_proj.Projection()
        proj.geod = PP.Geod(ellps="WGS84")
        proj.is_enu = False       

        if geom.type == "Polygon":
            proj.model = HG.EnuModel(geom.exterior.coords[0])
        elif geom.type == "LineString" or geom.type == "Point":
            proj.model = HG.EnuModel(geom.coords[0])
        else:
            raise Exception("Geometry_proj not supported geometry.")
                
        return proj

    def to_polygon(self, values):
        return SP.geometry.Polygon(values)

    def to_linestring(self, values):
        return SP.geometry.LineString(values)

    def to_point(self, value):
        return SP.geometry.Point(value)

    def geographic_to_polygon(self,geoms:[HG.GeographicLocation]):
        return self.to_polygon([[x.lon,x.lat,x.alt] for x in geoms])

    def geographic_to_linestring(self,geoms:[HG.GeographicLocation]):
        return self.to_linestring([[x.lon,x.lat,x.alt] for x in geoms])

    def geographic_to_point(self,geom:HG.GeographicLocation):
        return self.to_point([geom.lon,geom.lat,geom.alt])

    def to_geographic(self, geometry):
        if geometry.type == "Polygon":
            return [HG.GeographicLocation(*x) for x in geometry.exterior.coords] 
        return [HG.GeographicLocation(*x) for x in geometry.coords]

    def bounding_box(self, geometry):
        return SP.geometry.box(*geometry.bounds)

    def rotate(self, geometry, projection, angle):
        return SP.geometry.Polygon([projection.model.rotate(p, math.radians(angle)) for p in geometry.exterior.coords])

    def to_enu(self,geom, projection:Projection = None):
                
        if projection is None:        
            projection = self.create_projection(geom)
            
        projection.is_enu = True

        if geom.type == "Polygon":
            enu_list = [projection.model.to_enu(e) for e in geom.exterior.coords]
            projection.geometry = SP.geometry.Polygon(enu_list)
        
        elif geom.type == "LineString":
            enu_list = [projection.model.to_enu(e) for e in geom.coords]
            projection.geometry = SP.geometry.LineString(enu_list)
        
        elif geom.type == "Point":
            enu_point = projection.model.to_enu(*geom.coords)
            projection.geometry = SP.geometry.Point(enu_point)   
        else:
            raise Exception("Geometry_proj not supported geometry.")
        
        return projection
    
    def to_geodetic(self,geom, projection:Projection):
                
        if projection is None:        
            raise Exception("Geometry_proj::to_wgs requires a projection.")
        
        if not projection.is_enu:
            raise Exception("Geometry_proj::to_wgs projection should be in enu.")
        
        p = Geometry_proj.Projection()
        p.model = projection.model
        p.geod = projection.geod

        if geom.type == "Polygon":
            geo_list = [projection.model.to_geodetic(e) for e in geom.exterior.coords]
            p.geometry = SP.geometry.Polygon(geo_list)
            p.is_enu = False
        
        elif geom.type == "LineString":
            geo_list = [projection.model.to_geodetic(e) for e in geom.coords]
            p.geometry = SP.geometry.LineString(geo_list)
            p.is_enu = False
                    
        elif geom.type == "Point":
            geo_point = projection.model.to_geodetic(*geom.coords)
            p.geometry = SP.geometry.Point(geo_point)
            p.is_enu = False     
        else:
            raise Exception("Geometry_proj not supported geometry.")
        
        return p

    def split_linestring(self,geometry,max_length:float):
        projection = self.create_projection(geometry)
        projection.geometry = geometry        
        
        return [ x.geometry for x in self.split_linestring_proj(projection,max_length)]

    def split_linestring_proj(self,projection:Projection,max_length:float):

        proj = projection
        incomming_is_enu = proj.is_enu
        
        if not incomming_is_enu:
            proj = self.to_enu(proj.geometry)
        
        length = proj.geometry.length
        equal_distance = length / math.ceil(length / max_length)
        distances = NP.arange(0, length, equal_distance)
        points = [proj.geometry.interpolate(distance)
            for distance in distances] + [proj.geometry.boundary[1]]
        
        projections = []
        for x in range(len(distances) - 1):
            p = Geometry_proj.Projection()
            p.geometry = SP.geometry.LineString([points[x], points[x+1]])
            p.is_enu = proj.is_enu
            p.geod = proj.geod
            p.model = proj.model
            projections.append(p)
    
        if not incomming_is_enu:
            projections = [self.to_geodetic(proj.geometry,proj) for proj in projections]
                
        return projections

    def point_to_square(self,geometry,width:float):        
        projection = self.create_projection(geometry)
        projection.geometry = geometry        

        return self.point_square_proj(projection,width).geometry

    def point_square_proj(self,projection:Projection,width):
        s = width / 2.0
        square =  SP.geometry.Polygon([(s, s, 0), (-s, s, 0), (-s, -s, 0), (s, -s, 0)])
        projection.is_enu = True
        return self.to_geodetic(square,projection)

    def buffer(self, geometry, value):
        projection = self.create_projection(geometry)
        projection.geometry = geometry        

        return self.buffer_proj(projection, value).geometry

    def buffer_proj(self, projection, value):
        proj = projection
        incomming_is_enu = proj.is_enu
        
        if not incomming_is_enu:
            proj = self.to_enu(proj.geometry)

        if proj.geometry.type == "Polygon":
            z = proj.geometry.exterior.coords[0][2]
        elif proj.geometry.type == "LineString" or proj.geometry.type == "Point":
            z = proj.geometry.coords[0][2]
        else:
            raise Exception("Geometry_proj not supported geometry.")

        geom = SP.geometry.Polygon([(*p, z) for p in proj.geometry.buffer(value).exterior.coords])        
        return self.to_geodetic(geom,proj)          




import numpy as np
from horus_gis import SchemaProvider
from horus_geopandas import HorusGeoDataFrame

import geopandas as gpd
import pandas as pd

import os

try:
    from sklearn.neighbors import BallTree
except ModuleNotFoundError:
    print("Install module 'sklearn' to run this example.")
    exit(1)

def dt_class_filter(database,index,indices):
    dt_class = database.at[index,'dt_class']
    return [idx for idx in indices if database.at[idx,'dt_class'] == dt_class]


# Algorithm
def cluster(radius,tree, row_idx,database, datapoints, filter):

    # we have seen this point before
    if database.at[row_idx,'clstr_id'] != 0:
        return

    # Get the rows that fall within
    query  = np.array([datapoints[row_idx]])
    ind = tree.query_radius(query, r=radius)
    ind = filter(database,row_idx,ind[0])

    relation = {idx:database.iloc[idx]['clstr_id'] for idx in ind}
    cids = {relation[k] for k in relation.keys()}

    cids.remove(0)

    length = len(cids)

    # all elements will receive the row_idx as clstr_id
    if length == 0:
        for k in relation:
            database.at[k,'clstr_id'] = row_idx
            database.at[k, 'clstr_conf'] = 1.0

    # add instace to the found cluster
    if len(cids) == 1:
        new_id  = next(iter(cids))
        for k in relation.keys():
            if database.at[k,'clstr_id'] != new_id:
                database.at[k,'clstr_id'] = new_id
                database.at[k, 'clstr_conf'] = 1.0

    if len(cids) > 2:
        print("Edge case")
        exit(1)


### Read Data ####
sp = SchemaProvider()
schema = sp.merge(sp.single_measurement(),sp.clustering())
database = HorusGeoDataFrame(schema)

# provide a map of the new 'fields' with their default value
# missing fields will cause a schema exception
single_measurement_path = "output/single_measurement.shp"
if not os.path.exists(single_measurement_path):
    print(f"File '{single_measurement_path}' not found.\nRun 'horus_media_examples.spherical_camera_single_measurement' first.")
    exit(1)

database.append_file(single_measurement_path, default_values={'clstr_id':0,'clstr_conf':0.0})

# database has performed schema checks, and we will only update the cluster fields
# hence we can use the Geopandas dataframe directly
dataframe = database.dataframe



# Convert datapoints
datapoints = []

for index, row in dataframe.iterrows(): # Looping over all points
    datapoints.append(np.deg2rad([row['geometry'].y,row['geometry'].x]))


# Setup the Search tree
tree = BallTree(datapoints, metric='haversine')

# Adjust parameters
r_meter = 6371000.0
radius = 3.0 * (1.0/ r_meter)


# go fish
for index, row in dataframe.iterrows():
    cluster(radius,tree,index,dataframe,datapoints,dt_class_filter)


# Create output directory
output_dir = os.path.join(os.getcwd(), 'output')

if not os.path.exists(output_dir):
    os.makedirs(output_dir)

database.write_shapefile(os.path.join(output_dir, "single_measurement_clusters.shp"))




# Copyright(C) 2021 Horus View and Explore B.V.


import os
import sys
import csv
import logging
import itertools

from horus_db import Frames, Frame
from horus_media import Client, ImageRequestBuilder, ImageProvider, Size, Direction, \
    ComputationRequestBuilder, ComputationProvider
from horus_gis import PositionVector, Geographic

from . import util


# Comand line input handling
parser = util.create_argument_parser()
util.add_database_arguments(parser)
util.add_server_arguments(parser)
util.add_size_argument(parser, (1024, 1024))
parser.add_argument("-hf", "--horizontal-fov", type=float, default=(90),
                    help="horizontal field of view size in degrees")
parser.add_argument("-o", "--overlap", type=float, default=(20),
                    help="horizontal overlap in degrees")
parser.add_argument("--path", type=str, help="path to output location")
parser.add_argument("--limit", metavar=("NUMBER"), type=int,
                    help="maximum number of frames")
parser.add_argument("-r", "--recording", metavar="ID",
                    nargs='*', type=int, help="recording id")


args = parser.parse_args()

# This example shows how to request a spherical image
output_path = args.path

if output_path is None:
    logging.error(f"Specify an output path using '--path'")
    exit()

# Utility


def get_position_vector(frame, size, direction, fov):
    request_builder = ComputationRequestBuilder(frame.recordingid, frame.uuid)
    request = client.fetch(request_builder.build(size, direction, fov, 0, 0))
    result = computation_provider.fetch(request)
    return PositionVector(*result.values())


connection = util.get_connection(args)
client = util.get_client(args)

frames = Frames(connection)
image_provider = ImageProvider()
computation_provider = ComputationProvider()

# Output parameters
size = Size(args.size[0], args.size[1])  # defaults to (1024px, 1024px)
horizontal_fov = args.horizontal_fov  # defaults to 90 deg
overlap = args.overlap  # 20deg
recording_id = tuple(args.recording) if args.recording != None else None

displacement = horizontal_fov - overlap

directions = {
    "right": Direction(yaw=displacement, pitch=0),
    "center": Direction(yaw=0, pitch=0),
    "left": Direction(yaw=-displacement, pitch=0),
}

csv_header = [
    "File",
    "Index",
    "Altitude",
    "Azimuth",
    "Latitude",
    "Longitude",
    "Pitch",
    "Roll",
    "Stamp"
]


# Open output csv file
with open(os.path.join(output_path, 'output.csv'), 'w', newline='') as csvfile:
    writer = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL)
    writer.writerow(csv_header)

    # Get frames
    results = Frame.query(frames, recordingid=recording_id, order_by="index",)
    if args.limit:
        results = itertools.islice(results, args.limit)
    for frame in results:
        if frame is None:
            print("No frames!")
            exit()

        print(frame)

        # Get the image
        # Set parameters
        request_builder = ImageRequestBuilder(frame.recordingid, frame.uuid)
        for direction_id, direction in directions.items():
            request = client.fetch(request_builder.build_spherical(
                size, direction, horizontal_fov))
            result = image_provider.fetch(request)

            # Save the file
            filename = "{}_{}.jpg".format(frame.index, direction_id)

            with open(os.path.join(output_path, filename), 'wb') as image_file:
                image_file.write(result.image.getvalue())
                result.image.close()

            # Compute accurate position with lever arm.
            position = get_position_vector(
                frame, size, direction, horizontal_fov)

            writer.writerow([
                filename,
                frame.index,
                position.alt,
                frame.azimuth,
                position.lat,
                position.lon,
                frame.pitch,
                frame.roll,
                frame.stamp.isoformat()])



from horus_analytics import Clustering_analysis
import os


single_measurement_path = "output/single_measurement_clusters.shp"
if not os.path.exists(single_measurement_path):
    print(f"File '{single_measurement_path}' not found.\nRun 'horus_media_examples.spherical_camera_single_measurement' first.")
    exit(1)

cluster_locations_path = "output/cluster_locations.shp"
if not os.path.exists(cluster_locations_path):
    print(f"File '{cluster_locations_path}' not found.\nRun 'horus_media_examples.triangulate_spherical_camera_singlemeasurement_clusters' first.")
    exit(1)

ca = Clustering_analysis()
ca.cluster_data_layer(single_measurement_path)
ca.cluster_location(cluster_locations_path)

ca.analyse_cluster_entries()
ca.analyse_cluster()

ca.write_package("output/clustering_analysis.gpkg")


"""Horus Gis

Utilities for GIS ENU to and from Geodetic conversions.
"""
# Copyright(C) 2019, 2020 Horus View and Explore B.V.

import numpy
import pymap3d
import math
import argparse
from typing import NamedTuple
from ast import literal_eval
from scipy.spatial.transform import Rotation
from enum import Enum

def angle_between(v1, v2, up=numpy.array((0, 0, 1))):
    """ Returns the angle in radians between vectors 'v1' and 'v2'

    The angle is oriented according to the up vector.
    """

    angle = numpy.arccos(numpy.clip(numpy.dot(v1 / numpy.linalg.norm(v1), \
                v2 / numpy.linalg.norm(v2)), -1.0, 1.0))
    if numpy.dot(numpy.cross(v1, v2), up) < 0:
        return -angle
    return angle


def xyz_to_yxz(p):
    """Swaps x and y iterating over a 3D vector"""
    yield p[1]
    yield p[0]
    yield p[2]


def geoPointParser(string):
    try:
        value = literal_eval(f"({string})")
        value = tuple(map(lambda x: float(x), value))
        if len(value) == 3:
            return value
    except:
        pass
    raise argparse.ArgumentTypeError(
        "'{string}' is not a valid location")


class EnuModel:
    """East North Up spatial of reference

    Conversions between ENU and Geodetic."""

    east = numpy.array((1, 0, 0))
    north = numpy.array((0, 1, 0))
    up = numpy.array((0, 0, 1))

    def __init__(self, geodeticPoint):
        self.geodeticPoint = geodeticPoint

    def to_enu(self, geodeticPoint):
        return numpy.array(pymap3d.geodetic2enu(
            *xyz_to_yxz(geodeticPoint), *xyz_to_yxz(self.geodeticPoint)))

    def to_geodetic(self, point):
        """
        ENU to Azimuth, Elevation, Range

        Parameters
        ----------

        point : collection
            ENU point (meters, meters, meters)

        Results
        -------
            WGS84 point (lon, lat, alt)
        """
        return numpy.array([x for x in xyz_to_yxz(
            pymap3d.enu2geodetic(*point, *xyz_to_yxz(self.geodeticPoint)))])

    def rotate(self, point, angle):
        r = Rotation.from_euler('z', angle, degrees=True)
        return r.apply(point)

    def to_orientation(self, angle):
        # should use altitude ?
        return self.rotate(EnuModel.north, angle)

    def angle_between(self, v1, v2):
        return math.degrees(angle_between(v1, v2, [0, 0, 0]))

    def get_heading(self, point):
        return math.degrees(angle_between(EnuModel.north, point))
    
    def get_direction(self, point):
        yaw = (math.degrees(math.atan2(*point[:2])) + 360) % 360
        r = Rotation.from_euler('z', yaw, degrees=True)
        pitch = 90 - math.degrees(math.atan2(*(r.apply(point)[1:3])))
        return [yaw, pitch]


class CameraModel(EnuModel):
    def __init__(self, origin, heading):
        super(CameraModel, self).__init__(origin)
        self.orientation = self.to_orientation(-heading)

    @staticmethod
    def with_leverarms(origin, heading, leverarms):
        enu = EnuModel(origin)
        r = Rotation.from_euler('z', -heading + 90.0, degrees=True)
        return CameraModel(enu.to_geodetic(r.apply(leverarms)), heading)

    def look_at(self, location):
        return self.get_direction(self.to_enu(location))

    def look_at_angle(self, location):
        enu_location = self.to_enu(location)
        return math.degrees(angle_between(self.orientation, enu_location))


class GeographicLocation():
    """GeographicLocation longitude,latitude,altitude

    Geographic location refers to a position on the Earth. 
    Your absolute geographic location is defined by two coordinates,
    longitude and latitude. In the case of imagery, the altitude is added/required
    to be able to georeference pixels from this imagery.

    The reference system is that in which the recording has been stored in the database,
    which is usually EPSG Projection 4326 - WGS 84. 
    """
    lat: float
    lon: float
    alt: float

    def __init__(self, lon: float, lat: float, alt: float):
        self.lon = lon
        self.lat = lat
        self.alt = alt

    @staticmethod
    def from_tuple(tuple_lon_lat_alt):
        return GeographicLocation(
            tuple_lon_lat_alt[0],
            tuple_lon_lat_alt[1],
            tuple_lon_lat_alt[2])


class RelativeLocation():
    """RelativeLocation, east, north, up

        The relative location in meters using the ENU coordinate system.
    """
    east: float
    north: float
    up: float

    def __init__(self, east: float, north: float, up: float):
        self.east = east
        self.north = north
        self.up = up

    def location(self):
        return [self.east, self.north, self.up]

class PositionVector(NamedTuple):
    lat: float
    lon: float
    alt: float
    yaw: float
    pitch: float

class Geographic:
    @staticmethod
    def __normalize(v):
        norm = numpy.linalg.norm(v)
        if norm == 0: 
            return v
        return v / norm

    @staticmethod
    def __get_line(lat, lon, alt, bearing, pitch):
        p = pymap3d.geodetic2ecef(lat, lon, alt)

        d = numpy.array([math.cos(math.radians(pitch)) * math.sin(math.radians(bearing)),
                        math.cos(math.radians(pitch)) * math.cos(math.radians(bearing)),
                        math.sin(math.radians(pitch))])

        m = numpy.array([[-1 * math.sin(math.radians(lon)), 
                        -1 * math.sin(math.radians(lat)) * math.cos(math.radians(lon)), 
                        math.cos(math.radians(lat)) * math.cos(math.radians(lon))], 
                        [math.cos(math.radians(lon)), 
                        -1 * math.sin(math.radians(lat)) * math.sin(math.radians(lon)), 
                        math.cos(math.radians(lat)) * math.sin(math.radians(lon))],
                        [0, math.cos(math.radians(lat)), math.sin(math.radians(lat))]])

        d = m.dot(d)
        d = Geographic.__normalize(d)

        return p, d

    @staticmethod
    def __get_point(lines):
        m = numpy.zeros([3,3])
        left = numpy.zeros([3,3])
        right = numpy.zeros([3])

        for origin, direction in lines:
            m = numpy.identity(3) - \
                numpy.column_stack([numpy.array([direction * direction[0]]).T, 
                numpy.array([direction * direction[1]]).T, 
                numpy.array([direction * direction[2]]).T])
            left += m
            right += m.dot(origin)

        output = numpy.linalg.inv(left).dot(right)

        return numpy.array(pymap3d.ecef2geodetic(*output))
    
    @staticmethod
    def triangulate(pos_vectors):
        """ Triangulate position based on at least 3 position vectors """
        lines = []
        for i in pos_vectors:
            lines.append(Geographic.__get_line(*i))
        return Geographic.__get_point(lines)



class SchemaProvider:
    ### Provides database schemas ###
    class Geometry:
        class Type(Enum):
            POINT_2D = 1
            POINT_3D = 2
        ### placeholder geometry class ###
        type:Type
        def __init__(self,type:Type):
            self.type = type

        def __repr__(self):
            if self.type == self.Type.POINT_2D:
                return 'Point'
            if self.type == self.Type.POINT_3D:
                return '3D Point'
                
    class AutoIncrement:
        def __init__(self):
            pass

    class Field:
        name:str
        description:str
        type

        def __init__(self, name:str, description,type):
            self.name = name
            self.description = description
            self.type = type

        def __repr__(self):
            cls = type(self)
            return f"{cls.__name__}({self.name}, {getattr(self.type, '__name__', self.type)})"

    class Schema:
        fields=[]
        name:str

        def __init__(self, name: str):
            self.name = name

    @staticmethod
    def merge(schema_a, schema_b):
        schema = SchemaProvider.Schema(schema_a.name + " / " + schema_b.name)
        schema.fields = schema_a.fields + schema_b.fields
        return schema

    @staticmethod
    def single_measurement():
        schema = SchemaProvider.Schema("Single measurement schema")
        schema.fields =  [
            SchemaProvider.Field("geometry","Center of the detection projected on the surface.",
                                 SchemaProvider.Geometry(SchemaProvider.Geometry.Type.POINT_3D)),

            #Frame fields (viewpoint information)
            SchemaProvider.Field("rec_id", "The id of the recording.", int),
            SchemaProvider.Field("frame_idx", "The index of the frame.", int),
            SchemaProvider.Field("azimuth", "Heading/Azimuth (true north) of movement of the camera.", float),
            SchemaProvider.Field("usr_id", "An unique id for this record.", int),


            #SchemaProvider.Field("camera_geom", "Position of the camera.",
            #                     SchemaProvider.Geometry(SchemaProvider.Geometry.Type.POINT_3D)),
            #Camera fields (rendering parameters)
            SchemaProvider.Field("cam_lat", "The latitude of the camera.", float),
            SchemaProvider.Field("cam_lon", "The longitude of the camera.", float),
            SchemaProvider.Field("cam_alt", "The altitude of the camera.", float),
            SchemaProvider.Field("cam_fov", "The field of view of the spherical camera.",float),
            SchemaProvider.Field("cam_yaw", "The yaw of the spherical camera.",float),
            SchemaProvider.Field("cam_pitch", "The pitch of the spherical camera.",float),
            SchemaProvider.Field("cam_width", "The width of the spherical camera.",int),
            SchemaProvider.Field("cam_height", "The height of the spherical camera.",int),

            # Detection fields
            SchemaProvider.Field("dt_class", "The detection class/type ID.",int),
            SchemaProvider.Field("dt_name", "The detection name.",str),
            SchemaProvider.Field("dt_x", "The highest left pixel of the detection.",int),
            SchemaProvider.Field("dt_y", "The highest top pixel of the detection.",int),
            SchemaProvider.Field("dt_width", "The width of the detection.",int),
            SchemaProvider.Field("dt_height", "The height of the detection.",int),
            SchemaProvider.Field("dt_conf", "The confidence.",float),
            SchemaProvider.Field("dt_dist", "The surface distance in meters to the detection.",float),
            SchemaProvider.Field("dt_px_x", "The pixel x coordinate of the detection(could be centroid x).", int),
            SchemaProvider.Field("dt_px_y", "The pixel y coordinate of the detection (could be centroid y).", int),
            SchemaProvider.Field("dt_yaw", "The geographical yaw of the camera wrt the detection.", float),
            SchemaProvider.Field("dt_pitch", "The geographical pitch of the camera wrt the detection.", float),
            # 
            SchemaProvider.Field("surf_px_x","The pixel x coordinate of the surface(could be centroid x).",int),
            SchemaProvider.Field("surf_px_y","The pixel y coordinate of the surface (could be centroid y).",int),
            SchemaProvider.Field("surf_yaw","The geographical yaw of the camera wrt the surface.",float),
            SchemaProvider.Field("surf_pitch","The geographical pitch of the camera wrt the surface.",float)

        ]

        return schema

    @staticmethod
    def clustering():
        schema = SchemaProvider.Schema("Clustering schema")
        schema.fields = [
            SchemaProvider.Field("clstr_id", "The cluster id of the cluster.", int),
            SchemaProvider.Field("clstr_conf", "Clustering confidence.", float),
        ]

        return schema

    def geometry_3dpoint(self):
        schema = SchemaProvider.Schema("Geometry 3dPoint")
        schema.fields = [
            SchemaProvider.Field("geometry", "Basic 3D Point geometry.",
                                 SchemaProvider.Geometry(SchemaProvider.Geometry.Type.POINT_3D))
        ]

        return schema


# Copyright(C) 2022 Horus View and Explore B.V.


import os
import logging

from horus_db import Frames, Frame
from horus_media import ImageRequestBuilder, ImageProvider, Mode, Scales

from . import util

# Comand line input handling
parser = util.create_argument_parser()
util.add_database_arguments(parser)
util.add_server_arguments(parser)
parser.add_argument("-t", "--time",  metavar=("START", "END"), nargs=2, type=str,
                    help="time interval to query. E.g. '2022-02-13T14:30:00' '2022-02-13T15:00:00'")
parser.add_argument("--path", type=str, help="path to output location")
parser.add_argument("--limit", metavar=("NUMBER"), type=int, default=100,
                    help="maximum number of frames")
parser.add_argument("-r", "--recording", metavar="ID",
                    nargs='*', type=int, help="recording id")


args = parser.parse_args()

# This example shows how to request a spherical image
output_path = args.path

if args.time is None or len(args.time) != 2:
    logging.error(f"Invalid time interval '{args.time}'")
    exit()

if output_path is None:
    logging.error(f"Specify an output path using '--path'")
    exit()

connection = util.get_connection(args)
client = util.get_client(args)

frames = Frames(connection)
image_provider = ImageProvider()

# Input parameters
recording_id = tuple(args.recording) if args.recording != None else None


# Get frames
results = Frame.query(frames,
                      # tuple of timestamp strings (begin, end)in ISO format
                      time_interval=args.time,
                      # recording id (optional)
                      recordingid=recording_id,
                      order_by="index",
                      limit=args.limit,
                      )

for frame in results:
    if frame is None:
        print("No frames!")
        exit()

    print(frame, "at", frame.stamp)

    # Get the image
    request_builder = ImageRequestBuilder(frame.recordingid, frame.uuid)
    request = client.fetch(request_builder.build(
        Mode.panoramic, Scales.Px_1024))
    result = image_provider.fetch(request)

    # Save the file
    filename = "{}_{}_{}.jpg".format(
        frame.recordingid, frame.index, frame.stamp)

    with open(os.path.join(output_path, filename), 'wb') as image_file:
        image_file.write(result.image.getvalue())
        result.image.close()



# Copyright(C) 2020 Horus View and Explore B.V.

import psycopg2
import os

from horus_db import Frames, Frame, Recordings, Recording
from horus_media import Client, ImageRequestBuilder, ImageProvider, Mode, Direction, Size
from horus_camera import SphericalCamera
from horus_gis import GeographicLocation

from . import util


# Create output directory
output_dir = os.path.join(os.getcwd(), 'output')

if not os.path.exists(output_dir):
    os.makedirs(output_dir)

# This example shows how to create snapshots based on geometry


# Comand line input handling
parser = util.create_argument_parser()
util.add_database_arguments(parser)
util.add_server_arguments(parser)

parser.add_argument("--path", type=str, help="path to output location")

args = parser.parse_args()

# This example shows how to request a spherical image
output_path = args.path

connection = util.get_connection(args)
client = util.get_client(args)
frames = Frames(connection)
image_provider = ImageProvider()


# Step 1. create and configure spherical camera
sp_camera = SphericalCamera()
sp_camera.set_network_client(client)




def take_shaphost(nr,recording:Recording, frame:Frame, positions):
    sp_camera.set_frame(recording, frame)
    size = sp_camera.look_at_all(positions,1024)

    spherical_image = sp_camera.crop_to_geometry(sp_camera.acquire(size), positions, True)

    with open(output_dir + '/snapshot'+ str(nr) + '.jpeg', 'wb') as image_file:
        image_file.write(spherical_image.get_image().getvalue())
        spherical_image.get_image().close()




snapshot_1 = ["DemoData\Rotterdam360\Ladybug5plus" , 16, [
    GeographicLocation (4.4883159319642543,51.908492027690912, 53.361715),
    GeographicLocation (4.4884184967537415,51.908456244872745, 53.361715),
    GeographicLocation (4.48847610048205,51.908519406825874, 53.361715),
    GeographicLocation (4.4883735356838885,51.90855518966265, 53.361715)
    ]]


snapshot_2 = ["DemoData\Rotterdam360\Ladybug5plus" , 981, [
    GeographicLocation (4.4643636161038245,51.910632575408812,47.090053),
    GeographicLocation (4.4644063962677194,51.910582531617663,47.090053),
    GeographicLocation ( 4.46443096611172,51.910590565679584,47.090053),
    GeographicLocation (4.464388185974328,51.9106406094716,47.090053)
    ]]


snapshot_3 = ["DemoData\Rotterdam360\Ladybug5plus" , 981, [
    GeographicLocation (4.4644173663344944,51.910644188474507,47.090053),
    GeographicLocation (4.464431966331305,51.910622670221294,47.090053),
    GeographicLocation (4.4644493531869944,51.910627182612672,47.090053),
    GeographicLocation (4.4644347531983941,51.910648700866034,47.090053)
]]

snapshots = [snapshot_1,snapshot_2,snapshot_3]



recordings = Recordings(connection)
cursor = recordings.all()
recording = Recording(cursor)
while recording is not None:
    if snapshot_1[0] in recording.directory:
        recordings.get_setup(recording)

        nr = 0
        for snapshot in snapshots:
            results = Frame.query(frames, recordingid=recording.id,
                              index=snapshot[1],)
            frame = next(results)
            take_shaphost(nr,recording,frame,snapshot[2])
            nr += 1

    # next recording
    recording = Recording(cursor)


print("done")


# Copyright(C) 2021 Horus View and Explore B.V.

import psycopg2
import sys
import itertools

from horus_media import Size
from horus_camera import SphericalCamera, Pixel, SphericalImage, GeoReferencedPixel
from horus_db import Frames, Recordings, Frame, Recording
from horus_gis import SchemaProvider, GeographicLocation

from . import util


#------- If geopandas & pandora are installed
database = None
try:
    from horus_geopandas import HorusGeoDataFrame
    import geopandas as gpd

    sp = SchemaProvider()
    database = HorusGeoDataFrame(sp.single_measurement())

except ModuleNotFoundError:
    pass

def fill_record(record,grp: GeoReferencedPixel, camera:SphericalCamera, image:SphericalImage):
    print("TEST--->", )
    print("TEST--->",  image.get_geo_location_camera().lat)

    # update geometry
    record["geometry"] = geopandas.points_from_xy(x=[grp.geo_location.lon], y=[grp.geo_location.lat])
    # Record
    record["rec_id"] = grp.recording_id
    record["frame_idx"] = grp.frame_index
    #
    record["cam_fov"] = camera.h_fov.value
    record["cam_yaw"] = camera.yaw
    record["cam_pitch"] = camera.pitch
    record["cam_width"] = image.get_resolution().width
    record["cam_height"] = image.get_resolution().height
    record["cam_lat"] = image.get_geo_location_camera().lat
    record["cam_lon"] = image.get_geo_location_camera().lon
    record["cam_alt"] = image.get_geo_location_camera().alt
    #
    record["dt_class"] = 1234
    record["dt_name"] = "pole"
    record["dt_x"] = 12
    record["dt_y"] = 12
    record["dt_width"] = 13
    record["dt_height"] = 13
    record["dt_conf"] = 0.9
    record["dt_surf_x"] = 13
    record["dt_surf_y"] = 0
    record["dt_dist"] = grp.distance
    #
    record["vp_px_x"] = grp.pixel_location.col
    record["vp_px_y"] = grp.pixel_location.row
    record["vp_yaw"] = grp.viewing_parameters.yaw
    record["vp_pitch"] = grp.viewing_parameters.pitch



# ----------    Overview  --------------
#
# This example shows how a series --minimum of 3-- of geo referenced pixels (GeoReferencedPixel)
# obtained from a spherical image are used to calculate an accurate GeographicLocation for a specific label.
#
# This is done in several steps:
#
# 0: Configuration
# 1: Create the SphericalCamera which we will use to acquire SphericalImages
# 2: Get some frames from the City of Rotterdam
# 3: Simulate pixel/points of interest within the images, (these where picked by hand)
# 4: Iterate over the labelled points of intererst per frame and apply the camera's yaw,pitch to look at that point.
# 5: Obtain the GeoReferencedPixel at out point of interest from step 4
# 6: Triangulate the labelled GeoReferencedPixels (minimum of 3 are required)
# 7: Write the geo-data to standard out in CSV format, which can be visually be verified by programs such a QGis

# Step 0
# Configuration part
parser = util.create_argument_parser()
util.add_database_arguments(parser)
util.add_server_arguments(parser)

args = parser.parse_args()

connection = util.get_connection(args)
client = util.get_client(args)
recordings = Recordings(connection)

# Select the demo city of Rotterdam
recording = next(Recording.query(
    recordings, directory_like="Rotterdam360\\\\Ladybug5plus"))
recordings.get_setup(recording)
print(recording, " -> ", recording.directory)
print(recording.setup)

# Step 1. create and configure spherical camera
sp_camera = SphericalCamera()
sp_camera.set_network_client(client)

# Step 2. Get a recorded frame and place the camera onto that 'frame'
frames = Frames(connection)
results = Frame.query(frames, recordingid=recording.id,
                      index=210, order_by="index",)

# step 3. do some sight seeing..
# This sections hardcodes looking at points of interest, which
# normally would come from image processing or AI tooling.


class Poi():
    """Describes and labels a point of interest within the image """
    pixel: Pixel
    name: str
    yaw: float
    pitch: float

    def __init__(self, pixel, name, yaw, pitch) -> None:
        self.name = name
        self.pitch = pitch
        self.yaw = yaw
        self.pixel = pixel


# The list that describes:
# [ frame index -> [Points of interests] ]
frames_of_interest = [
    [210, [
        Poi(Pixel(287, 353), "p1", 20, -30),
        Poi(Pixel(209, 284), "p2", 20, -30),
        Poi(Pixel(213, 595), "p3", -40, -30)
    ]],
    [212, [
        Poi(Pixel(465, 590), "p1", 20, -30),
        Poi(Pixel(215, 343), "p2", 20, -30),
        Poi(Pixel(225, 604), "p3", -40, -30)
    ],
    ],
    [214, [
        Poi(Pixel(379, 613), "p1", 120, -30),
        Poi(Pixel(234, 303), "p2", 30, -30),
        Poi(Pixel(244, 542), "p3", -40, -30)
    ]
    ]
]

# Step 4,
# That actual processing of the provided data
# Creating a set of labelled geo referenced pixels [GeoReferencedPixel]
# which we will use to triangulate in step 6 for a more accurate location
list_of_images = []

for foi in frames_of_interest:
    results = Frame.query(frames, recordingid=recording.id,
                          index=foi[0], order_by="index",)
    frame = next(results)
    if frame is None:
        print("No frames!")
        exit()

    print("Got frame:", frame, frame.get_location())
    sp_camera.set_frame(recording, frame)
    print("camera_height", sp_camera.height)

    for poi in foi[1]:

        # -- obtain database
        record = None
        if database is not None:
            record = database.new_frame()

        sp_camera.set_horizontal_fov(90)
        sp_camera.set_yaw(poi.yaw)
        sp_camera.set_pitch(poi.pitch)

        spherical_image = sp_camera.acquire(Size(800, 800))

        with open('frame' + str(frame.index) + '_' + poi.name + '.jpeg', 'wb') as image_file:
            image_file.write(spherical_image.get_image().getvalue())
            spherical_image.get_image().close()

        # Step 5
        # obtain a GeoReferencedPixel
        grp: GeoReferencedPixel = spherical_image.project_pixel_on_ground_surface(
            poi.pixel)
        grp.name = poi.name

        # you can store multiple GeoReferencedPixels with a different label
        spherical_image.store_geo_referenced_pixel(grp)

        list_of_images.append(spherical_image)

        if record is not None:
            fill_record(record,grp,sp_camera,spherical_image)
            database.add_frame(record)

    if database:
        record = database.new_frame()

# Step 6
# Triangulate the labels
p1_tri: GeographicLocation = SphericalImage.triangulate("p1", list_of_images)
p2_tri: GeographicLocation = SphericalImage.triangulate("p2", list_of_images)
p3_tri: GeographicLocation = SphericalImage.triangulate("p3", list_of_images)

# Step 7 verify the results
# Print all the locations and info (in csv format)
print("")
print("latitude longitude altitude name frame recording row col")

# print  GeoReferencedPixel's
for img in list_of_images:
    for name in img.geo_referenced_pixels:
        grp: GeoReferencedPixel = img.geo_referenced_pixels[name]
        print(grp.geo_location.lat, grp.geo_location.lon, grp.geo_location.alt, grp.name,
              grp.frame_index, grp.recording_id, grp.pixel_location.row, grp.pixel_location.col)

print(p1_tri.lat, p1_tri.lon, p1_tri.alt, "p1_tr", -1, -1, -1, -1)
print(p2_tri.lat, p2_tri.lon, p2_tri.alt, "p2_tr", -1, -1, -1, -1)
print(p3_tri.lat, p3_tri.lon, p3_tri.alt, "p3_tr", -1, -1, -1, -1)



# Copyright(C) 2021 Horus View and Explore B.V.

import sys
from horus_db import Frames, Recordings, Frame, Recording
from horus_media import Grid, ImageProvider, ImageRequestBuilder, Mode, Scales
from . import util
from random import sample
import logging


# Command line input handling
parser = util.create_argument_parser()
util.add_database_arguments(parser)
util.add_server_arguments(parser)

args = parser.parse_args()

connection = util.get_connection(args)
client = util.get_client(args)
image_provider = ImageProvider()

recordings = list(Recording.iter(Recordings(connection).all()))
frames = Frames(connection)
grid = Grid()

streamHandler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
streamHandler.setFormatter(formatter)

logger = logging.getLogger()
logger.addHandler(streamHandler)
logger.info("Started")
logger.setLevel(logging.INFO)

while True:
    recording = sample(recordings, 1)[0]
    cursor = frames.query(recordingid=recording.id)

    frame = Frame(cursor)

    while frame is not None:
        request_builder = ImageRequestBuilder(frame.recordingid, frame.uuid)
        requests = client.fetch_all(request_builder.build(
            Mode.panoramic, Scales.Px_2048, section) for section in grid)
        logging.info(f"{frame} {request_builder}")
        frame = Frame(cursor)





"""Horus Media"""
# Copyright(C) 2019, 2020 Horus View and Explore B.V.


from dataclasses import dataclass
from PIL import Image
from itertools import chain
from enum import Enum

import urllib.parse
import io
import math
import time
import json
import logging
import http.client


@dataclass
class Point:
    x: float = 0
    y: float = 0

    def __iter__(self):
        yield self.x
        yield self.y

    def __getitem__(self, ii):
        """Get a list item"""
        if ii == 0:
            return self.x
        if ii == 1:
            return self.y

    def __len__(self):
        return 2


@dataclass
class Rect(Point):
    width: float = 0
    height: float = 0


@dataclass
class Box:
    min: Point
    max: Point

    @classmethod
    def create(cls, center, width=0, height=0):
        if not isinstance(center, Point):
            center = Point(*center)
        x_interval = (center.x + width/2, center.x - width/2)
        y_interval = (center.y + height/2, center.y - height/2)
        return Box(Point(min(x_interval), min(y_interval)), Point(max(x_interval), max(y_interval)))


@dataclass(frozen=True)
class Section:
    x: int
    y: int
    ax: float
    ay: float
    index: int


class Grid:
    """ Grid  8x4 (c x r) """

    def __init__(self, w_min=-180, w_max=180, h_min=-90, h_max=90, r=4, c=8):
        assert w_min < w_max
        assert h_min < h_max
        self.rows = r
        self.cols = c
        self.section_height = (h_max - h_min)/r
        self.section_width = (w_max - w_min)/c
        self.__map = {}
        for index in range(r * c):
            x = index % c
            y = (r-1) - index // c
            ax = x*self.section_width - w_max
            ay = y*self.section_height - h_max
            self.__map[index] = Section(x, y, ax, ay, index)

    def __iter__(self):
        """ Returns the Iterator object """
        return (section for section in self.__map.values())

    def __str__(self):
        return str(self.__map)

    def __getitem__(self, ii):
        """Get a list item"""
        return self.__map[ii]

    def __len__(self):
        """List length"""
        return len(self.__map)

    def __contains__(self, value):
        if isinstance(value, Section):
            return value.index in self.__map
        return value in self.__map

    @dataclass(frozen=True)
    class Comparator:
        min: float
        max: float
        size: float

        def __contains__(self, value):
            return self.min < (value + self.size) and value < self.max

    def filter(self, fov=None, w_min=-180, w_max=180, h_min=-90, h_max=90):
        """
        fov is a Box with min, max points
        """
        if fov:
            w_min = fov.min.x
            h_min = fov.min.y
            w_max = fov.max.x
            h_max = fov.max.y
        assert w_min < w_max
        assert h_min < h_max
        assert h_min >= -90
        assert h_max <= 90

        w_min_wrapped = (w_min % 360) - (360 if w_min < 0 else 0)
        if w_min_wrapped < -180:
            w_min_wrapped = 360 + w_min_wrapped

        w_max_wrapped = (w_max % 360) - (360 if w_max < 0 else 0)
        if w_max_wrapped > 180:
            w_max_wrapped = -360 + w_max_wrapped

        h = self.Comparator(h_min,  h_max, self.section_height)
        if w_min_wrapped < w_max_wrapped:
            w = self.Comparator(
                w_min_wrapped,  w_max_wrapped, self.section_width)
            for section in self.__map.values():
                if section.ay in h and section.ax in w:
                    yield section
        elif w_min_wrapped > w_max_wrapped:
            assert w_max_wrapped >= -180
            assert w_min_wrapped <= 180
            w_left = self.Comparator(-180,  w_max_wrapped, self.section_width)
            w_right = self.Comparator(w_min_wrapped,  180, self.section_width)
            for section in self.__map.values():
                if section.ay in h and (section.ax in w_left or section.ax in w_right):
                    yield section


@dataclass(frozen=True)
class Scale:
    id: int
    size: int


@dataclass(frozen=True)
class Scales:
    Px_256 = Scale(1, 256)
    Px_512 = Scale(2, 512)
    Px_1024 = Scale(3, 1024)
    Px_2048 = Scale(4, 2048)
    Px_4096 = Scale(5, 4096)

    __SCALES = (
        Px_256,
        Px_512,
        Px_1024,
        Px_2048,
        Px_4096,
    )

    @classmethod
    def from_size(cls, size):
        """ Returns the closest round-up scale that contains the requested size.
        Returns None if no applicable scale is found.
        """
        result = cls.__SCALES[-1]
        for scale in reversed(cls.__SCALES):
            if scale.size < size:
                return result
            result = scale
        return result

class Mode(Enum):
    panoramic = 0
    spherical = 1
    orthographic = 2
    geographic = 3

    def __str__(self):
        return self.name

@dataclass(frozen=True)
class Size:
    width: int
    height: int

@dataclass(frozen=True)
class Direction:
    yaw: float
    pitch: float

@dataclass(frozen=True)
class Geometry:
    scale: int
    width: float
    height: float
    distance: float
    shift: float
    altitude: float

class Client:
    def __init__(self, url="http://localhost:5050/web/", timeout=4):
        self.url = url
        self.__parsed_url = urllib.parse.urlparse(url)
        self.__connection = http.client.HTTPConnection(
            self.__parsed_url.hostname, self.__parsed_url.port, timeout=timeout)
        self.__connection.connect()
        self.attempts = 5
        self.attempts_interval = 3 #seconds

    def fetch(self, request):
        request.url = urllib.parse.urljoin(
            self.__parsed_url.path, request.resource)
        self.__connection.request("GET", request.url)
        attempts = self.attempts
        while attempts > 0:
            attempts -= 1
            try:
                response = self.__connection.getresponse()
                attempts = 0
            except Exception as exception:
                logging.error(f'{exception}. Requesting "{request.url}".')
                logging.error(f'New attempt in {self.attempts_interval}s .')
                time.sleep(self.attempts_interval)

        request.response = response
        result = response.read()
        if type(request.local_file) == str:
            with open(request.local_file, "wb") as file:
                file.write(result)
        request.set_result(result)
        return request

    def fetch_all(self, requests):
        jobs = []
        for request in requests:
            self.fetch(request)
            jobs.append(request)

        return jobs


class ImageRequest:
    def __init__(self, builder, resource, mode=None, scale=None, section=None, size=None, direction=None, fov=None, cams = None, geometry=None):
        self.local_file = None
        self.builder = builder
        self.resource = resource
        self.mode = mode
        self.scale = scale
        self.section = section
        self.size = size
        self.direction = direction
        self.fov = fov
        self.cams = cams
        self.geometry = geometry
        self.url = None
        self.response = None
        self.__result = None

    def set_result(self, value):
        self.__result = value

    def result(self):
        return self.__result
    def __repr__(self):
        cls = type(self)
        return f"{cls.__name__}({self.url})"


class ImageRequestBuilder:
    def __init__(self, recording, frame):
        self.path_template = None
        self.recording = recording
        self.frame = frame
        self.__resource = f"./images/{recording}/{frame}"

    def __repr__(self):
        cls = type(self)
        return f"{cls.__name__}({self.__resource})"

    def build_spherical(self, size=None, direction=None, vof=None, cams=None):
        return self.build(Mode.spherical, None, None, size, direction, vof, cams, None)

    def build_orthographic(self, size, geometry):
        return self.build(Mode.orthographic, None, None, size, None, None, None, geometry)

    def build_geographic(self, size, direction=None, vof=None, x=None, y=None):
        if x and y:
            geometry = Geometry(0, x, y, 0, 0, 0)
        else:
            geometry = None
        return self.build(Mode.geographic, None, None, size, direction, vof, None, geometry)

    def build(self, mode=None, scale=None, section=None, size=None, direction=None, fov=None, cams=None, geometry=None):
        data = {}
        if mode is not None:
            data["mode"] = mode
        if scale is not None:
            data["scale"] = scale.id
        if section is not None:
            data["section"] = section.index
        if size is not None:
            data["size"] = str(size.width) + 'x' + str(size.height)
        if direction is not None:
            data["yaw"] = direction.yaw
            data["pitch"] = direction.pitch
        if fov is not None:
            data["hor_fov"] = fov
        if cams is not None:
            data["cams"] = cams
        if geometry is not None:
            data["scale"] = geometry.scale
            data["geom_width"] = geometry.width
            data["geom_height"] = geometry.height
            data["geom_dist"] = geometry.distance
            data["geom_shift"] = geometry.shift
            if geometry.altitude is not None:
                data["alti_next"] = geometry.altitude
        url_values = urllib.parse.urlencode(data)
        url = urllib.parse.urljoin(self.__resource, "?" + url_values)
        request = ImageRequest(self, url, mode, scale, section, size, direction, fov, cams, geometry)
        if self.path_template:
            request.local_file = self.path_template.format(
                recording=self.recording, frame=self.frame, mode=mode if mode else "", 
                scale=scale.id if scale else "", section=section.index if section else "")
        return request

class ImageProvider:
    @dataclass(frozen=True)
    class Result:
        image: io.BytesIO
        fov: Rect
        w: int
        h: int

        def to_pixel_coordinates(self, point):
            """
            the components of thepoint represent two angles (in Grid space)
            """
            px, py = point
            py = (py/py*(py % 90) if py != 0 else 0) + 90
            dy = 180 - (self.fov.y/self.fov.y*(self.fov.y % 90) if self.fov.y != 0 else 0)
            y = (1 + (py - dy)/self.fov.height)*self.h

            px = (px/px*(px % 180) if px != 0 else 0) + 180
            dx = (self.fov.x/self.fov.x*(self.fov.x % 180) if self.fov.x != 0 else 0) + 180
            x = (px - dx)/self.fov.width*self.w

            return Point(math.floor(x),math.floor(y))

    def __init__(self, grid=Grid()):
        self.grid = grid

    def fetch(self, image_request, w = None, h = None):
        return self.Result(io.BytesIO(image_request.result()), 
            Rect(0, 0, w, h) if w and h else Rect(0, 0, 1, 1), w if w else 0, h if h else 0)

    def combine(self, image_requests, w, h):
        rows = set()
        cols = set()
        for req in image_requests:
            if req.section in self.grid:
                rows.add(req.section.y)
                cols.add(req.section.x)

        row_map = self.set_to_map(sorted(rows))
        col_map = self.set_to_map(self.wrap(sorted(cols)))
        row_index_shift = len(rows) - 1

        stitched = Image.new('RGB', (w * len(col_map), h * len(row_map)))
        fov = Rect(math.inf, math.inf, 0, 0)
        fov.width = len(col_map) * self.grid.section_width
        fov.height = len(row_map) * self.grid.section_height

        for req in image_requests:
            r = row_map[req.section.y]
            c = col_map[req.section.x]
            fov.y = min(req.section.ay, fov.y)
            fov.x = min(req.section.ax, fov.x)

            try:
                image = Image.open(io.BytesIO(req.result()))
                stitched.paste(image, (c * w, (row_index_shift - r) * h))
            except Exception as exception:
                logging.error(f"{exception}. Stitching section {req.section} from {req.url}")
        image = io.BytesIO();
        stitched.save(image, format='jpeg', quality=95)
        return self.Result(image, fov, stitched.width, stitched.height)

    @classmethod
    def wrap(cls, cols):
        if len(cols) > 0:
            prev = cols[0]
            list_1 = []
            list_2 = []
            for x in cols:
                if x - prev > 1:
                    list_1, list_2 = (list_2, list_1)
                list_1.append(x)
                prev = x

            return [x for x in chain(list_1, list_2)]
        return cols

    @classmethod
    def set_to_map(cls, my_set):
        size = len(my_set)
        if size > 0:
            mx = min(my_set)
            mp = range(size+1)
            return dict(zip(my_set, mp))
        return {}

class ComputationRequest:
    def __init__(self, builder, resource, size=None, direction=None, fov=None, x=None, y=None, direction0=None):
        self.local_file = None
        self.builder = builder
        self.resource = resource
        self.size = size
        self.direction = direction
        self.fov = fov
        self.x = x
        self.y = y
        self.direction0 = direction0
        self.url = None
        self.response = None
        self.__result = None

    def set_result(self, value):
        self.__result = value

    def result(self):
        return self.__result

class ComputationRequestBuilder:
    def __init__(self, method, frame = None):
        self.path_template = None
        self.method = method
        self.frame = frame
        self.__resource = f"./computation/{method}"
        if frame != None:
            self.__resource += f"/{frame}"

    def build(self, size=None, direction=None, fov=None, x=None, y=None, direction0=None):
        data = {}
        if size is not None:
            data["size"] = str(size.width) + 'x' + str(size.height)
        if direction is not None:
            data["yaw"] = direction.yaw
            data["pitch"] = direction.pitch
        else:
            data["yaw"] = 0
            data["pitch"] = 0
        if fov is not None:
            data["hor_fov"] = fov
        if x is not None:
            data["x"] = x
        if y is not None:
            data["y"] = y
        if direction0 is not None:
            data["yaw0"] = direction0.yaw
            data["pitch0"] = direction0.pitch
        url_values = urllib.parse.urlencode(data)
        url = urllib.parse.urljoin(self.__resource, "?" + url_values)
        return ComputationRequest(self, url, size, direction, fov, x, y, direction0)

class ComputationProvider:
    @dataclass(frozen=True)
    class Result:
        data: io.BytesIO

    def fetch(self, computation_request):
        try:
            result = self.Result(io.BytesIO(
                computation_request.result())).data.getvalue()
            return json.loads(result)
        except json.decoder.JSONDecodeError as error:
            return {"error": error}



import sys
import horus_spatialite
from os import path

sqlite_file = "path_to_your_sqlite_database"

if len(sys.argv) > 1:
    sqlite_file = sys.argv[1]

if not path.exists(sqlite_file):
    print("Database file: [", sqlite_file, "] could not be found.")
    exit()

db = horus_spatialite.Spatialite(sqlite_file)

for arg in sys.argv[2:]:
    db.blob_contains_geometry(arg) 

db.open()
db.resolve()
db.show_info()

# Iterate over data
for row in db.query():
    for fieldname, info in db.get_field_names_map().items():
        # print(row[info.idx])
        pass

db.close()



# Copyright(C) 2021, 2022 Horus View and Explore B.V.
import copy
import os
from pprint import pprint


import json
from horus_media import Size
from horus_camera import SphericalCamera, Pixel, SphericalImage, GeoReferencedPixel, ViewParameterizedPixel
from horus_db import Frames, Recordings, Frame, Recording
from horus_gis import SchemaProvider
from . import util
from horus_geopandas import HorusGeoDataFrame
import geopandas as gpd

def fill_record(record,frame,grp: GeoReferencedPixel, det_vp: ViewParameterizedPixel,camera:SphericalCamera, image:SphericalImage, detection):

    # update geometry
    record["geometry"] = gpd.points_from_xy(x=[grp.geo_location.lon], y=[grp.geo_location.lat])
    # Record
    record["rec_id"] = grp.recording_id
    record["frame_idx"] = grp.frame_index
    record["azimuth"] = frame.azimuth
    #
    record["cam_fov"] = camera.h_fov.value
    record["cam_yaw"] = camera.yaw
    record["cam_pitch"] = camera.pitch
    record["cam_width"] = image.get_resolution().width
    record["cam_height"] = image.get_resolution().height
    record["cam_lat"] = image.get_geo_location_camera().lat
    record["cam_lon"] = image.get_geo_location_camera().lon
    record["cam_alt"] = image.get_geo_location_camera().alt
    #
    record["dt_class"] = detection.classification
    record["dt_name"] =  detection.name
    record["dt_x"] = detection.detection_pixel.col
    record["dt_y"] = detection.detection_pixel.row
    record["dt_width"] = detection.detection_size.width
    record["dt_height"] = detection.detection_size.height
    record["dt_conf"] = detection.confidence
    record["dt_dist"] = grp.distance

    # viewing parameters of the detection
    record["dt_px_x"] = det_vp.pixel_location.col
    record["dt_px_y"] = det_vp.pixel_location.row
    record["dt_yaw"] = det_vp.viewing_parameters.yaw
    record["dt_pitch"] = det_vp.viewing_parameters.pitch

    # viewing parameters of the surface projection
    record["surf_px_x"] = grp.pixel_location.col
    record["surf_px_y"] = grp.pixel_location.row
    record["surf_yaw"] = grp.viewing_parameters.yaw
    record["surf_pitch"] = grp.viewing_parameters.pitch



# ----------    Overview  --------------
#
# This example shows how a series --minimum of 3-- of geo referenced pixels (GeoReferencedPixel)
# obtained from a spherical image are used to calculate an accurate GeographicLocation for a specific label.
#
# This is done in several steps:
#
# 0: Configuration
# 1: Create the SphericalCamera which we will use to acquire SphericalImages
# 2: Get some frames from the City of Rotterdam
# 3: Simulate pixel/points of interest within the images, (these where picked by hand)
# 4: Iterate over the labelled points of intererst per frame and apply the camera's yaw,pitch to look at that point.
# 5: Obtain the GeoReferencedPixel at out point of interest from step 4
# 6: Triangulate the labelled GeoReferencedPixels (minimum of 3 are required)
# 7: Write the geo-data to standard out in CSV format, which can be visually be verified by programs such a QGis

# Step 0
# Configuration part
parser = util.create_argument_parser()
util.add_database_arguments(parser)
util.add_server_arguments(parser)

args = parser.parse_args()

connection = util.get_connection(args)
client = util.get_client(args)
recordings = Recordings(connection)


# Opening JSON file
f = open('input/spherical_camera_single_measurement.json')
data = json.load(f)

# Select the demo city of Rotterdam
recording = next(Recording.query(
    recordings, directory_like=data["recording_name"]))
recordings.get_setup(recording)
print(recording, " -> ", recording.directory)
print(recording.setup)




# Step 1. create and configure spherical camera
sp_camera = SphericalCamera()
sp_camera.set_network_client(client)

# Step 2. Get a recorded frame and place the camera onto that 'frame'
frames = Frames(connection)
results = Frame.query(frames, recordingid=recording.id,
                      index=210, order_by="index",)

# step 3. do some sight seeing..
# This sections hardcodes looking at points of interest, which
# normally would come from image processing or AI tooling.

sp = SchemaProvider()
database = HorusGeoDataFrame(sp.single_measurement())


class Camera:
    yaw: float
    pitch: float
    cam_size: Size
    cam_hor_fov:float

    def __init__(self):
        self.yaw = 0
        self.pitch = -10
        self.cam_size = Size(800,800)
        self.cam_hor_fov = 90

    def load(self,data):
        if "width" in data:
            self.cam_size = Size(int(data["width"]),self.cam_size.height)
        if "height" in data:
            self.cam_size = Size(self.cam_size.width,int(data["height"]))
        if "yaw" in data:
            self.yaw = float(data["yaw"])
        if "pitch" in data:
            self.pitch = float(data["pitch"])
        if "horizontal_fov" in data:
            self.cam_hor_fov = float(data["horizontal_fov"])


    def clone(self):
        return copy.deepcopy(self)

class Detection:
    """Describes and a Detection / Point of interest within the image """
    surface_pixel: Pixel
    detection_pixel: Pixel
    detection_size: Size
    classification :int
    name:str
    confidence : float
    camera: Camera

    def __init__(self,camera) -> None:
        self.camera = camera

    def load(self,dt):
        self.surface_pixel = Pixel(int(dt["surface_pixel"]["r"]),int(dt["surface_pixel"]["c"]))
        self.detection_pixel = Pixel(int(dt["detection_pixel"]["r"]), int(dt["detection_pixel"]["c"]))
        self.detection_size = Size(int(dt["detection_size"]["width"]), int(dt["detection_size"]["height"]))

        self.name = dt["name"]
        self.confidence = float(dt["confidence"])
        self.classification = int(dt["classification"])

        if "camera" in dt:
            self.camera.load(dt["camera"])







# Step 4,
# That actual processing of the provided data
# Creating a set of labelled geo referenced pixels [GeoReferencedPixel]
# which we will use to triangulate in step 6 for a more accurate location
list_of_images = []

camera = Camera()
camera.load(data["camera"])

detections_per_frame = data["detections_per_frame"]

# Create output directory
output_dir = os.path.join(os.getcwd(), 'output')

if not os.path.exists(output_dir):
    os.makedirs(output_dir)


for frame_id in detections_per_frame:
    results = Frame.query(frames, recordingid=recording.id,
                          index=frame_id, order_by="index",)
    frame = next(results)
    if frame is None:
        print("No frames!")
        exit()

    print("Got frame:", frame, frame.get_location())
    sp_camera.set_frame(recording, frame)
    print("camera_height", sp_camera.height)


    for detection_id in detections_per_frame[frame_id]:

        detection = Detection(camera.clone())
        detection.load(detections_per_frame[frame_id][detection_id])

        # -- obtain database frame
        record = database.new_frame()

        # -- set the camera
        sp_camera.set_horizontal_fov(detection.camera.cam_hor_fov)
        sp_camera.set_yaw(detection.camera.yaw)
        sp_camera.set_pitch(detection.camera.pitch)

        # -- acquire
        spherical_image = sp_camera.acquire(detection.camera.cam_size)

        # with open(output_dir + '/frame' + str(frame.index) + '_' + detection.name + '.jpeg', 'wb') as image_file:
        #    image_file.write(spherical_image.get_image().getvalue())
        #    spherical_image.get_image().close()

        # Step 5
        # obtain a GeoReferencedPixel on the surface
        grp: GeoReferencedPixel = spherical_image.project_pixel_on_ground_surface(
            detection.surface_pixel)
        grp.name = detection.name

        # step 6
        # obtain the ViewParameterizedPixel of the actual detection
        detection_parms = spherical_image.get_view_parameterized_pixel(detection.detection_pixel);

        fill_record(record,frame,grp,detection_parms,sp_camera,spherical_image, detection)
        record["usr_id"] = database.dataframe.shape[0]
        database.add_frame(record)



database.write_shapefile(os.path.join(output_dir, "single_measurement.shp"))




"""Horus database"""
# Copyright(C) 2019, 2020, 2021 Horus View and Explore B.V.

import logging
from typing import NamedTuple

class Table:
    attributes = {}
    repr_attributes = ["id"]

    def __new__(cls, cursor):
        self = cls._instance = object.__new__(cls)
        self.__result = cursor.fetchone()
        if not self.__result:
            return None

        for key, value in zip(cursor.description, self.__result):
            setattr(self, key.name, value)
        return self

    def __getattr__(self, name):
        if name in type(self).attributes:
            return self.__dict__[type(self).attributes[name]]
        raise Exception(f"No attribute {name}")

    @classmethod
    def query(cls, model, **kwargs):
        cursor = model.query(**kwargs)
        return cls.iter(cursor)

    @classmethod
    def iter(cls, cursor):
        item = cls(cursor)
        while item:
            yield item
            item = cls(cursor)

    def __repr__(self):
        cls = type(self)
        attrs = ', '.join('{}={}'.format(attr, getattr(self, attr))
                          for attr in type(self).repr_attributes)
        return f"{cls.__name__}({attrs})"


class LeverArm(NamedTuple):
    x: float
    y: float
    z: float


class RecordingSetup(Table):
    attributes = {
        "camera_height": "cameraHeight",
        "lever_arm_x": "leverArmX",
        "lever_arm_y": "leverArmY",
        "lever_arm_z": "leverArmZ",
        "id": "recording_id",
    }
    repr_attributes = ["id", "camera_height", "lever_arm"]

    @property
    def lever_arm(self):
        return LeverArm(self.leverArmX,
                        self.leverArmY,
                        self.leverArmZ)


class Recording(Table):
    attributes = {
        "directory": "recordingdirectory",
        "bounding_box": "boundingbox",
        "file_format": "fileformat",
    }
    repr_attributes = ["id", "boundingbox"]
    setup: RecordingSetup = None


class Frame(Table):
    attributes = {
        "heading": "azimuth",
        "timestamp": "stamp",
        "uuid": "guid"
    }
    repr_attributes = ["id", "recordingid", "index"]

    def get_location(self):
        return (
            self.longitude,
            self.latitude,
            self.altitude)


class RecordingSetups:
    def __init__(self, connection):
        self.__connection = connection

    def all(self):
        cursor = self.__connection.cursor()
        cursor.execute('SELECT * FROM "MoviePlayer_recordingsetup"')
        return cursor

    def get(self, id):
        cursor = self.__connection.cursor()
        cursor.execute("""SELECT "cameraHeight", "leverArmX", "leverArmY", "leverArmZ", "recording_id"
FROM "MoviePlayer_recordingsetup" WHERE recording_id = %s;""", (id,))
        return cursor


class Recordings:
    def __init__(self, connection):
        self.__connection = connection

    def get(self, id):
        cursor = self.__connection.cursor()
        cursor.execute("""SELECT id, recordingdirectory, boundingbox, fileformat
FROM recordings WHERE id = %s;""", (id,))
        return cursor

    def get_setup(self, recording):
        cursor = self.__connection.cursor()
        cursor.execute("""SELECT "cameraHeight", "leverArmX", "leverArmY", "leverArmZ", "recording_id"
FROM "MoviePlayer_recordingsetup" WHERE recording_id = %s;""", (recording.id,))
        recording.setup = RecordingSetup(cursor)

    def all(self):
        cursor = self.__connection.cursor()
        cursor.execute("""SELECT id, recordingdirectory, boundingbox, fileformat
FROM recordings""")
        return cursor

    def query(self, **kwargs):
        select_clause = {
            "id",
            "recordingdirectory",
            "boundingbox",
            "fileformat",
        }
        where_clause = []
        orderby_clause = []
        params = []
        tail = []

        for arg, value in kwargs.items():
            if value == None:
                continue
            if arg in select_clause:
                if type(value) != tuple:
                    value = (value,)
                where_clause.append(f"{arg}" + " IN %s")
                params.append(value)
                continue
            if arg == "directory_like":
                where_clause.append("recordingdirectory LIKE %s")
                params.append(f"%{value}%")
                continue
            if arg == "limit":
                tail.append("limit %s")
                params.append(value)
                if "offset" in kwargs:
                    tail.append("offset %s")
                    params.append(kwargs["offset"])
                continue
            if arg == "offset":
                continue
            if arg == "order_by":
                if type(value) != tuple:
                    value = (value,)
                orderby_clause += value
                continue

            logging.warning(
                f'Recordings query unknown argument "{arg}" skipped')

        sql = "SELECT " + ", ".join(select_clause) + " FROM recordings"

        if len(where_clause) > 0:
            sql += " WHERE " + " AND ".join(where_clause)

        if len(orderby_clause) > 0:
            orderby_clause.append("id")
            sql += " ORDER BY  " + ", ".join(orderby_clause)

        if len(tail) > 0:
            sql += " " + " ".join(tail) + ";"

        cursor = self.__connection.cursor()
        cursor.execute(sql, params)
        return cursor


class Frames:
    def __init__(self, connection):
        self.__connection = connection

    def query(self, **kwargs):
        select_clause = {
            "id",
            "recordingid",
            "guid",
            "latitude",
            "longitude",
            "altitude",
            "roll",
            "pitch",
            "azimuth",
            "stamp",
            "index"
        }
        where_clause = []
        orderby_clause = []
        params = []
        tail = []

        for arg, value in kwargs.items():
            if value == None:
                continue
            if arg == "within":
                point, distance = value
                st_point = f"ST_SetSRID(ST_Point({point[0]}, {point[1]}), 4326)"
                select_clause.add(
                    f"ST_Distance(geom::geography, {st_point}::geography) as distance")
                orderby_clause.append("distance")
                where_clause.append(
                    f"ST_DWithin(geom::geography, {st_point}::geography, {distance})")
                continue
            if arg == "distance":
                point, operator, distance = value
                st_point = f"ST_SetSRID(ST_Point({point[0]}, {point[1]}), 4326)"
                where_clause.append(
                    f"ST_Distance(geom::geography, {st_point}::geography) {operator}")
                params.append(distance)
                continue
            if arg == "time_interval":
                start, end = value
                orderby_clause.append("stamp")
                where_clause.append(f"stamp BETWEEN '{start}' AND '{end}'")
                continue
            if arg in select_clause:
                if type(value) != tuple:
                    value = (value,)
                where_clause.append(f"{arg}" + " IN %s")
                params.append(value)
                continue
            if arg == "limit":
                tail.append("limit %s")
                params.append(value)
                if "offset" in kwargs:
                    tail.append("offset %s")
                    params.append(kwargs["offset"])
                continue
            if arg == "offset":
                continue
            if arg == "order_by":
                if type(value) != tuple:
                    value = (value,)
                orderby_clause += value
                continue

            logging.warning(f'Frames query unknown argument "{arg}" skipped')

        sql = "SELECT " + ", ".join(select_clause) + " FROM frames"

        if len(where_clause) > 0:
            sql += " WHERE " + " AND ".join(where_clause)

        orderby_clause.append("id")
        if len(orderby_clause) > 0:
            sql += " ORDER BY  " + ", ".join(orderby_clause)

        if len(tail) > 0:
            sql += " " + " ".join(tail) + ";"

        cursor = self.__connection.cursor()
        cursor.execute(sql, params)
        return cursor


def Iterator(cursor):
    item = cursor.fetchone()
    while item:
        yield item
        item = cursor.fetchone()


def print_all_results(cursor):
    """
    Iterates over a cursor, printing each results.
    The cursor will be consumed.
    """
    print(cursor.rowcount, "results")
    for result in Iterator(cursor):
        print("\n")
        for key, value in zip((column.name for column in cursor.description), result):
            print("  ", key, ":", value)



# Copyright(C) 2021 Horus View and Explore B.V.
"""Horus Media Server strategy example

Example utility
"""

import psycopg2
import argparse
import logging

from horus_media import Client, Geometry


class read_arguments_from_file(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        file_name = None
        if option_string == "--db-file":
            file_name = namespace.db_file

        if option_string == "--s-file":
            file_name = namespace.s_file

        if file_name != None:
            f = open(file_name, "r")
            parser.parse_args(f.read().split(), namespace)

def create_argument_parser():
    return argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

def add_database_arguments(parser):
    # db
    parser.add_argument("--db-name", type=str,
                        default="HorusWebMoviePlayer", help="the database name")
    parser.add_argument("--db-user", type=str, default="postgres",
                        help="database user name used to authenticate")
    parser.add_argument("--db-password", type=str,
                        help="database password used to authenticate")
    parser.add_argument("--db-host", type=str, default="localhost",
                        help="database database host address")
    parser.add_argument("--db-port", type=int, default=5432,
                        help="database connection port number")

    parser.add_argument('--db-file', help="database configuration file", type=str,default="input/db.conf",action=read_arguments_from_file,nargs=0)


def add_server_arguments(parser):
    parser.add_argument("-s", "--server",  metavar="URL", type=str, default="http://localhost:5050/web/",
                        help="Horus Media Server endpoint")
    parser.add_argument("-st", "--server-timeout",  metavar="SECONDS", type=int, default=4,
                        help="Horus Media Server timeout in seconds")
    parser.add_argument("-sa", "--server-attempts",  metavar="ATTEMPTS", type=int, default=5,
                        help="Horus Media Server number of attempts on connection failure")

    parser.add_argument('--s-file', help="server configuration file",type=str, default="input/server.conf", action=read_arguments_from_file, nargs=0)


def add_size_argument(parser, default=(1024, 1024)):
    parser.add_argument("-S", "--size",  metavar=("WIDTH", "HEIGHT"), nargs=2, type=int, default=default,
                        help="size of the image")


def add_geometry_arguments(parser):
    parser.add_argument("-gsc", "--geom-scale", type=int, default=400,
                        help="output scale in px/m (orthographic)")
    parser.add_argument("-gw", "--geom-width", type=float, default=6,
                        help="geometry width (orthographic)")
    parser.add_argument("-gh", "--geom-height", type=float, default=2,
                        help="geometry height (orthographic)")
    parser.add_argument("-gd", "--geom-dist", type=float, default=4,
                        help="geometry distance (orthographic)")
    parser.add_argument("-gs", "--geom-shift", type=float, default=0,
                        help="geometry shift (orthographic)")


def get_database_connection_string(args):
    db_params = [("host", args.db_host),
                 ("port", str(args.db_port)),
                 ("dbname", args.db_name),
                 ("user", args.db_user),
                 ("password", args.db_password),
                 ]

    return " ".join(
        map("=".join, filter(lambda x: x[1] != None, db_params)))


def get_connection(args):
    try:
        return psycopg2.connect(get_database_connection_string(args))
    except psycopg2.OperationalError as exception:
        logging.error(f"{exception} Connecting to database")
        exit()


def get_geometry(args, altitude_next=None):
    return Geometry(args.geom_scale, args.geom_width,
                    args.geom_height, args.geom_dist, args.geom_shift, altitude_next)


def get_client(args):
    try:
        client = Client(args.server, args.server_timeout)
        client.attempts = args.server_attempts
        return client
    except OSError as exception:
        logging.error(f"{exception}. Connecting to server {args.server}")
        exit(1)


def sample_script_header(name):
    if name == "__main__":
        print(f"This module is a sample script. It cannot run as an interactive example.")
        exit(1)



from horus_gis import SchemaProvider
import horus_gis as hg
from horus_geopandas import HorusGeoDataFrame

import pandas as pd
import geopandas as gpd
import fiona

from math import radians, cos, sin, asin, sqrt

def haversine(point1, point2 ): #lon1, lat1, lon2, lat2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)
    """

    lon1 = point1.x
    lat1 = point1.y
    lon2 = point2.x
    lat2 = point2.y

    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a))
    m = 6367000 * c
    return m



class Clustering_analysis:

    cluster_layer: HorusGeoDataFrame
    cluster_location_layer: HorusGeoDataFrame
    schema_provider:SchemaProvider

    layers:None

    def __init__(self):
        self.schema_provider = SchemaProvider()
        self.layers ={}

    def cluster_data_layer(self, filename, layer=None):
        schema = self.schema_provider.merge(
            self.schema_provider.single_measurement(),
            self.schema_provider.clustering())

        self.cluster_layer = HorusGeoDataFrame(schema)
        self.cluster_layer.append_file(filename,layer)

    def cluster_location(self, filename, layer=None):
        schema = self.schema_provider.merge(
            self.schema_provider.geometry_3dpoint(),
            self.schema_provider.clustering())
        self.cluster_location_layer = HorusGeoDataFrame(schema)
        self.cluster_location_layer.append_file(filename,layer)

    def analyse_cluster_entries(self):
        """
        Adds the cluster_entries_analysis layer with:
        distance: meter to triangulation
        """

        schema = SchemaProvider.Schema("Analysis cluster entries")
        schema.fields = [
            SchemaProvider.Field("geometry", "Row idx (observeration) geometry.",
                                 SchemaProvider.Geometry(SchemaProvider.Geometry.Type.POINT_3D)),
            SchemaProvider.Field("usr_id", "The id", int),
            SchemaProvider.Field("distance", "Distance between approximation and triangulation.", int),
            SchemaProvider.Field("angle", "Angle between camera movement and detected object.", float)
        ]
        database = HorusGeoDataFrame(schema)

        loc_df = self.cluster_location_layer.dataframe
        data_df = self.cluster_layer.dataframe
        for index, row in loc_df.iterrows():
            mask = data_df['clstr_id'] == row['clstr_id']
            entries = data_df.loc[mask]
            for point_i, point_r in entries.iterrows():
                record = database.new_frame()
                record['geometry'] = point_r['geometry']
                record['usr_id'] = point_r['usr_id']
                record['distance'] = haversine(row['geometry'],point_r['geometry'])
                record['angle'] = (point_r['dt_yaw'] - point_r['azimuth'])

                database.add_frame(record)

        self.layers["cluster_entries_analysis"] = database

    def analyse_cluster(self):
        """
        Adds the cluster_analysis layer with:
        distance: meter to triangulation
        """
        schema = SchemaProvider.Schema("Analysis clusters")
        schema.fields = [
            SchemaProvider.Field("geometry", "Cluster geometry.",
                                 SchemaProvider.Geometry(SchemaProvider.Geometry.Type.POINT_3D)),
            SchemaProvider.Field("clstr_pts", "Number of points used for clustering.", int)
        ]
        database = HorusGeoDataFrame(schema)

        loc_df = self.cluster_location_layer.dataframe
        data_df = self.cluster_layer.dataframe
        for index, row in loc_df.iterrows():
            mask = data_df['clstr_id'] == row['clstr_id']
            entries = data_df.loc[mask]

            record = database.new_frame()
            geoms = []
            for point_i, point_r in entries.iterrows():
                geoms.append(point_r['geometry'])
            record['clstr_pts'] = len(geoms)
            database.add_frame(record)

        self.layers["cluster_analysis"] = database

    def write_package(self, filename):
        self.layers["cluster_layer"] = self.cluster_layer
        self.layers["cluster_location_layer"] = self.cluster_location_layer
        for k, v in self.layers.items():
            v.write_geopackage(filename, layer=k)




import numpy as np
import os
import geopandas as gpd
import pandas as pd
import sys
from PIL import Image, ImageDraw
import io




from horus_gis import SchemaProvider
from horus_geopandas import HorusGeoDataFrame
from horus_camera import SphericalCamera
from horus_db import Frames, Recordings, Frame, Recording
from horus_media import Size
from . import util


# Step 0
# Configuration part
parser = util.create_argument_parser()
util.add_database_arguments(parser)
util.add_server_arguments(parser)
parser.add_argument("-f", "--frame", type=int,nargs=1)
parser.add_argument("-i", "--input", type=str,nargs=1)


args = parser.parse_args()




### Read Data ####
sp = SchemaProvider()
database = HorusGeoDataFrame(sp.single_measurement())

# provide a map of the new 'fields' with their default value
# missing fields will cause a schema exception
single_measurement_path = "output/single_measurement.shp"
if not os.path.exists(single_measurement_path):
    print(f"File '{single_measurement_path}' not found.")
    exit(1)

database.append_file(single_measurement_path)

## Prepare Work,
class Work:
    info = None
    indices:[]

worklist = {}

for index, row in database.dataframe.iterrows(): # Looping over all points
    if row['frame_idx'] == args.frame[0]:
        id = str(row['frame_idx']) + ":" + str(row['cam_width']) + "x" + str(row['cam_height'])
        id += "_[" + str(row['cam_fov']) + "][" + str(row['cam_yaw']) + "][" + str(row['cam_pitch'])+"]"

        if id not in worklist:
            worklist[id] = Work()
            worklist[id].info = row
            worklist[id].indices = [index]
        else:
            worklist[id].indices.append(index)

## Perform Work,



connection = util.get_connection(args)
client = util.get_client(args)
recordings = Recordings(connection)

# Step 1. create and configure spherical camera
sp_camera = SphericalCamera()
sp_camera.set_network_client(client)

# Step 2. Get a recorded frame and place the camera onto that 'frame'
frames = Frames(connection)

df = database.dataframe
cnt = 0
for workid in worklist:
   cnt+=1
   print(cnt,"/", len(worklist),workid)
   job = worklist[workid]

   recording = next(Recording.query(
      recordings, id=job.info["rec_id"]))
   recordings.get_setup(recording)

   results = Frame.query(frames,
       recordingid=job.info["rec_id"],
       index=job.info["frame_idx"], order_by="index",)

   frame = next(results)
   # -- set the camera
   sp_camera.set_frame(recording, frame)
   sp_camera.set_horizontal_fov(job.info['cam_fov'])
   sp_camera.set_yaw(job.info['cam_yaw'])
   sp_camera.set_pitch(job.info['cam_pitch'])

   # -- acquire
   spherical_image = sp_camera.acquire(Size(job.info["cam_width"],job.info["cam_height"]))


   # -- draw our findings

   data = spherical_image.get_image().getvalue()
   stream = io.BytesIO(data)
   img = Image.open(stream)


   draw = ImageDraw.Draw(img)
   spherical_image.get_image().close()

   for row in job.indices:
      #print(row,"/",len(job))
      record = df.loc[[row]]

      if record.iloc[0]['dt_class'] != -1:
         # pil upper left 0,0
         x0 = record.iloc[0]['dt_x']
         y0 = record.iloc[0]['dt_y']
         w = record.iloc[0]['dt_width']
         h = record.iloc[0]['dt_height']

         shape = [ x0, y0, x0 + w , y0 + h]
         draw.rectangle(shape, outline ="red")


         surf_x =  record.iloc[0]['surf_px_x']
         surf_y =  record.iloc[0]['surf_px_y']
         #print(surf_x,surf_y)
         draw.ellipse([surf_x-2, surf_y-2, surf_x+2, surf_y+2], fill = 'blue', outline ='blue')

      draw.text((surf_x + 4, surf_y), record.iloc[0]['dt_name'], fill=None, font=None, anchor=None, spacing=0, align="left")



      img.save('output/' + workid+'.jpeg')








"""Horus GeoPandas https://geopandas.org/en/stable/

Utilities and Classes build on top of GeoPandas
"""


from horus_gis import SchemaProvider
import horus_gis as hg

try:
    import pandas as pd
    import pandera as pa
    import fiona as fi
    from fiona.crs import from_epsg
    import geopandas as gpd
except ModuleNotFoundError as e:
    print(f"Install module '{e.name}' to use this module.")
    raise e

class HorusGeoDataFrame:

    schema: SchemaProvider.Schema
    dataframe: gpd.GeoDataFrame
    non_geom_schema: pa.DataFrameSchema
    fiona_schema:dict
    crs:str

    fields:None
    def __init__(self,schema:SchemaProvider.Schema):

        self.crs = "EPSG:4326"
        self.schema = schema

        # Convert schema's
        # https://pandera.readthedocs.io/en/latest/dataframe_schemas.html
        the_geom_field = False

        pan_schema = {}
        self.fiona_schema = {'geometry':None, 'properties':{}}
        self.fields=[]
        for field in self.schema.fields:
            is_geom:bool = type(field.type) is hg.SchemaProvider.Geometry

            if is_geom:
                pan_schema[field.name] = pa.Column(pa.engines.pandas_engine.Geometry)
                self.fiona_schema[field.name] = str(field.type)
            else:
                pan_schema[field.name] = pa.Column(field.type)
                self.fiona_schema['properties'][field.name] = \
                    list(fi.FIELD_TYPES_MAP.keys())[list(fi.FIELD_TYPES_MAP.values()).index(field.type)]

            self.fields.append(field.name)
        
        self.non_geom_schema = pa.DataFrameSchema(pan_schema,
            index=pa.Index(int),
            strict=True,
            coerce=True,)

        self.dataframe = gpd.GeoDataFrame(columns = self.fields)
        self.at = self.dataframe.at

    def new_frame(self, geom=None):
        if geom is None:
            geom = gpd.points_from_xy([0], [0],[0])
        return gpd.GeoDataFrame(columns = self.fields, geometry=geom,crs=self.crs)

    def add_frame(self, dataframe, validate=True):
        if validate:
            self.non_geom_schema.validate(dataframe)
        self.dataframe = pd.concat([self.dataframe, dataframe])

    def append_file(self,filename, layer=None,default_values={}):
        dataframe = gpd.read_file(filename,layer=layer, schema=self.fiona_schema)

        for dv in default_values.keys():
            dataframe[dv] = default_values[dv]

        self.non_geom_schema.validate(dataframe)

        self.dataframe = pd.concat([self.dataframe, dataframe])

    def write_shapefile(self, filename):
        self.dataframe.to_file(filename, crs=from_epsg(int(self.crs.split(':')[1])), 
        schema=self.fiona_schema)

    def write_geojson(self, filename):
        self.dataframe.to_file(filename, driver='GeoJSON', crs=from_epsg(int(self.crs.split(':')[1])), 
        schema=self.fiona_schema)
        
    def write_geopackage(self,filename,layer):
        self.dataframe.to_file(filename, layer=layer, driver="GPKG", crs=from_epsg(int(self.crs.split(':')[1])), 
        schema=self.fiona_schema)





# Copyright(C) 2021 Horus View and Explore B.V.


import os
import sys
import logging
import itertools

from horus_db import Frames, Frame
from horus_media import ImageRequestBuilder, ImageProvider, Size, ComputationProvider

from . import util


# Comand line input handling
parser = util.create_argument_parser()
util.add_database_arguments(parser)
util.add_server_arguments(parser)
util.add_size_argument(parser, (4096, 2048))
util.add_geometry_arguments(parser)

parser.add_argument("--path", type=str, help="path to output location")
parser.add_argument("--limit", metavar=("NUMBER"), type=int,
                    help="maximum number of frames")
parser.add_argument("-r", "--recording", metavar="ID",
                    nargs='*', type=int, help="recording id")

args = parser.parse_args()

# This example shows how to request orthographic images
output_path = args.path

if output_path is None:
    logging.error(f"Specify an output path using '--path'")
    exit()


connection = util.get_connection(args)
client = util.get_client(args)

frames = Frames(connection)
image_provider = ImageProvider()
computation_provider = ComputationProvider()

# Output parameters
size = Size(args.size[0], args.size[1])  # defaults to (4096px, 2048px)
recording_id = tuple(args.recording) if args.recording != None else None


# Get frames
results = Frame.query(frames, recordingid=recording_id, order_by="index",)
if args.limit:
    results = itertools.islice(results, args.limit)
for frame in results:
    if frame is None:
        print("No frames!")
        exit()

    print(frame)

    # Get the image
    # Set parameters
    request_builder = ImageRequestBuilder(frame.recordingid, frame.uuid)

    cursor = frames.query(recordingid=frame.recordingid, index=frame.index + 1)
    next_frame = Frame(cursor)
    if next_frame is not None:
        geometry = util.get_geometry(args, next_frame.altitude)
    else:
        geometry = util.get_geometry(args)

    request = client.fetch(
        request_builder.build_orthographic(size, geometry))

    result = image_provider.fetch(request)
    # Save the file
    filename = "orthographic_{}.tif".format(frame.index)
    with open(os.path.join(output_path, filename), 'wb') as image_file:
        image_file.write(result.image.getvalue())
        result.image.close()



# Copyright(C) 2021 Horus View and Explore B.V.

import psycopg2
import sys

from horus_db import Frames, Recordings, Frame, Recording
from . import util

# This example shows how to iterate over all the recordings

# Comand line input handling
parser = util.create_argument_parser()
util.add_database_arguments(parser)
util.add_server_arguments(parser)

args = parser.parse_args()

connection = util.get_connection(args)
recordings = Recordings(connection)

cursor = recordings.all()

recording = Recording(cursor)

while recording is not None:
    print(recording, " -> ", recording.directory)
    recording = Recording(cursor)



# Copyright(C) 2019, 2020 Horus View and Explore B.V.
"""Horus Media Server strategy example

Given a sign location
Assumption:
- the sign targeted face is oriented towards the driving direction
- We drive on the right-hand side of the road

Goal:
- narrow horizontal fov, centered on the sign
- the sign is visible head on within the grabbed images
- the images are within a max distance from the sign

In addition it exemplifies geo-query based acquisition of georeferenced orthographic projection images (geotiff)
"""

import psycopg2

from horus_gis import CameraModel, geoPointParser
from horus_db import Frames, Recordings, Recording, Frame, Iterator
from horus_media import Client, ImageRequestBuilder, ImageProvider, Grid, Scales, Point, Box, Mode, Size, Direction, Geometry

import argparse
import logging


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument("target", nargs='+', metavar="lon,lat,alt",
                    type=geoPointParser, help="target GPS location in decimal degrees")
parser.add_argument("-r", "--recording", metavar="ID",
                    nargs='*', type=int, help="recording id")
parser.add_argument("-s", "--server",  metavar="URL", type=str, default="http://localhost:5050/web/",
                    help="Horus Media Server endpoint")
parser.add_argument("-f", "--fov",  metavar=("HORIZONTAL", "VERTICAL"), nargs=2, type=float, default=(25, 90),
                    help="field of view size in degrees")
parser.add_argument("-m", "--mode",  type=str, default="panoramic", choices=[x.name for x in list(Mode)],
                    help="rendering mode")
parser.add_argument("-si", "--size",  metavar=("WIDTH", "HEIGHT"), nargs=2, type=int, default=(1024, 1024),
                    help="size of the image (spherical), size of the back-buffer (orthographic)")
parser.add_argument("-ci", "--clipping",  metavar=("MIN", "MAX"), type=float, nargs=2, default=(15, 100),
                    help="clipping interval in degrees")
parser.add_argument("-d", "--distance", type=float, default=10,
                    help="maximum distance from the target in meters")
parser.add_argument("--path", type=str, help="output location")
parser.add_argument("--attempts", metavar=("NUMBER"), type=int, help="maximum number of attempts", default=100)

# geom
parser.add_argument("-gsc", "--geom-scale", type=int, default=400,
                    help="output scale in px/m (orthographic)")
parser.add_argument("-gw", "--geom-width", type=float, default=6,
                    help="geometry width (orthographic)")
parser.add_argument("-gh", "--geom-height", type=float, default=2,
                    help="geometry height (orthographic)")
parser.add_argument("-gd", "--geom-dist", type=float, default=4,
                    help="geometry distance (orthographic)")
parser.add_argument("-gs", "--geom-shift", type=float, default=0,
                    help="geometry shift (orthographic)")

# db
parser.add_argument("--db-name", type=str,
                    default="HorusWebMoviePlayer", help="the database name")
parser.add_argument("--db-user", type=str, default="postgres",
                    help="database user name used to authenticate")
parser.add_argument("--db-password", type=str,
                    help="database password used to authenticate")
parser.add_argument("--db-host", type=str, default="localhost",
                    help="database database host address")
parser.add_argument("--db-port", type=int, default=5432,
                    help="database connection port number")

args = parser.parse_args()

temp_path = args.path
recording_id = tuple(args.recording) if args.recording != None else None
distance = args.distance
horizontal_fov = args.fov[0]
vertical_fov = args.fov[1]
mode = args.mode
size = Size(args.size[0], args.size[1])
clipping_interval = tuple(args.clipping)
geometry = Geometry(args.geom_scale, args.geom_width, args.geom_height, args.geom_dist, args.geom_shift, None)

db_params = [("host", args.db_host),
             ("port", str(args.db_port)),
             ("dbname", args.db_name),
             ("user", args.db_user),
             ("password", args.db_password),
             ]
try:
    connection_string = " ".join(
        map("=".join, filter(lambda x: x[1] != None, db_params)))
    connection = psycopg2.connect(connection_string)
except psycopg2.OperationalError as exception:
    logging.error(f"{exception} Connecting to database")
    exit()
try:
    client = Client(args.server)
except OSError as exception:
    logging.error(f"{exception}. Connecting to server {args.server}")
    exit()

frames = Frames(connection)
grid = Grid()
image_provider = ImageProvider(grid)


def compute_angle(frame, sign_location):
    camera_model = CameraModel(frame.get_location(), frame.heading)
    return - camera_model.look_at_angle(sign_location)


for location in args.target:
    print(f"Looking for {location}")
    results = []
    cursor = frames.query(within=(location, distance),
                          recordingid=recording_id, limit=1)
    frame = Frame(cursor)
    if not frame:
        logging.warning(
            f"Location {location} not found within {distance} meters")
        continue

    if recording_id:
        assert frame.recordingid in recording_id
    angle = compute_angle(frame, location)
    attempts = args.attempts
    while True:
        if (clipping_interval[0] <= angle <= clipping_interval[1]):
            requestBuilder = ImageRequestBuilder(frame.recordingid, frame.uuid)

            filename = None

            if mode == Mode.panoramic.name:
                fov = Box.create(center=(angle, 0),
                                width=horizontal_fov, height=vertical_fov)
                sections = grid.filter(fov=fov)
                requests = client.fetch_all(requestBuilder.build(
                    mode, Scales.Px_1024, section) for section in sections)
                result = image_provider.combine(
                    requests, Scales.Px_1024.size, Scales.Px_1024.size)
                if temp_path:
                    filename = temp_path + \
                        "stitched_{}_{}.jpg".format(
                            frame.recordingid,  frame.index)
                results.append({
                    "frame": frame,
                    "angle": angle,
                    "pixel_coordinate": result.to_pixel_coordinates((angle, 0)),
                    "stitch": result})

            if mode == Mode.spherical.name:
                request = client.fetch(requestBuilder.build_spherical(
                    size, Direction(angle, 0), horizontal_fov))
                result = image_provider.fetch(request, size.width, size.height)
                if temp_path:
                    filename = temp_path + "spherical_{}_{}.jpg".format(
                        frame.recordingid,  frame.index)
                results.append({
                    "frame": frame,
                    "angle": angle,
                    "pixel_coordinate": Point(size.width / 2, 0),
                    "stitch": result})

            if mode == Mode.orthographic.name:
                request = client.fetch(requestBuilder.build_orthographic(size, geometry))
                                                            
                result = image_provider.fetch(request)
                if temp_path:
                    filename = temp_path + "orthographic_{}_{}.tif".format(
                        frame.recordingid,  frame.index)
                results.append({
                    "frame": frame,
                    "angle": angle,
                    "pixel_coordinate": Point(0, 0),
                    "stitch": result})

            if filename:
                with open(filename, 'wb') as f:
                    f.write(result.image.getvalue())
                result.image.close()
        elif len(results) > 0:
            break
        elif attempts < 0:
            print("Exceeded maximum number of attempts.")
            break
        else:
            attempts = attempts - 1


        next_index = frame.index
        frame = Frame(cursor)
        if not frame:
            # request in batches of 40 frames using decreasing indexes
            indexes = tuple(range(next_index - 40, next_index))
            cursor = frames.query(
                index=indexes, order_by="index DESC", recordingid=recording_id)
            frame = Frame(cursor)
            if not frame:
                break
        angle = compute_angle(frame, location)

    for result in results:
        print()
        print("frame:", result["frame"])
        print("angle:", str(round(result["angle"], 2)), "deg")
        if "pixel_coordinate" in result:
            print("x coordinate:", str(result["pixel_coordinate"].x), "px")
        # print("fov:", result["stitch"].fov)


connection.close()



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='horus-media-client',

    version='0.9.1',

    description='Horus Media Server Client',

    long_description=long_description,

    long_description_content_type='text/markdown',

    url='https://github.com/horus-view-and-explore/horus-media-client',

    author_email="info@horus.nu",

    author='Horus View and Explore B.V.',

    license='MIT',

    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ],

    packages=find_packages(where='.'),

    python_requires='>=3.6, <4',

    install_requires=[
        "psycopg2>=2.8.4",
        "numpy>=1.17.4",
        "pymap3d>=2.1.0",
        "pillow>=6.2.1",
        "scipy>=1.3.3",
    ],

    extras_require={
        'dev': [],
        'test': [],
    },
)



from PIL import Image, ImageDraw
from io import BytesIO
import math
import io

from horus_db import Frame, Recording
from horus_media import Client, ImageRequestBuilder, ImageRequest, ImageProvider, Size, Direction, \
    ComputationRequestBuilder, ComputationProvider, Rect
from horus_gis import GeographicLocation, RelativeLocation, PositionVector, Geographic, CameraModel

horus_geometries_found = False

try:
    from horus_geometries import Geometry_proj
    horus_geometries_found = True
except:
    pass

class FieldOfView():
    value: float
    min: float
    max: float

    def __init__(self, value: float, min: float, max: float):
        self.value = value
        self.min = min
        self.max = max

class Pixel():
    row: int
    col: int

    def __init__(self, row: int,  col: int):
        self.row = row
        self.col = col


class ViewParameterizedPixel():
    """Parameters for georeferencing

    The ViewParameterizedPixel is a data structure that
    has all the information that is required for geo-referencing
    pixels. 

    ViewParameterizedPixels can be obtained from specific 
    Imagery that have geo-referencing capabilities, such as the SphericalImage.

    A set of 3 different ViewParameterizedPixels from 3 different locations
    can than be used for triangulation resulting in a GeographicLocation.

    Attributes:
        name                Label to identify the pixel
        pixel_location      The pixel location in pixel coordinates
        viewing_parameters
        recording_id        The id of the source recording
        frame_index         The index of the source frame in the source recording
        user_data           Dictionary to store optional user-defined data about this pixel
    """
    name: str
    pixel_location: Pixel
    viewing_parameters: PositionVector
    recording_id: int
    frame_index: int
    user_data: dict

    def __init__(self):
        self.user_data = {}


class GeoReferencedPixel(ViewParameterizedPixel):
    """Parameters for georeferencing with GeographicLocation.

    The GeoReferencedPixel is the same data structure as the 
    ViewParameterizedPixel with an additional GeographicLocation.

    A single ViewParameterizedPixel can be turned into a GeoReferencedPixel
    when additional constraints/assumptions are added to the Pixel/PositionVector.

    In the case of the SphericalImage, the function project_pixel_on_ground_surface 
    adds constraints on the altitude coordinate, thus allowing for an estimation
    on the GeographicLocation.

    The distance is the ground distance from camera to geo_location.
    """
    distance: float
    geo_location: GeographicLocation
    relative_loc: RelativeLocation

    def __init__(self, vpp: ViewParameterizedPixel):
        super(GeoReferencedPixel, self).__init__()
        self.name = vpp.name
        self.pixel_location = vpp.pixel_location
        self.viewing_parameters = vpp.viewing_parameters
        self.recording_id = vpp.recording_id
        self.frame_index = vpp.frame_index


"""Horus Camera

The camera
"""
# Copyright(C) 2019, 2020 Horus View and Explore B.V.


class Camera():
    """The base camera class.

    """
    h_fov: FieldOfView           # horizontal field of view
    v_fov: FieldOfView           # vertical field of view
    yaw: float                   # yaw of the camera ?immutable
    pitch: float                 # pitch of the camera ?immutable
    frame: Frame                 # current frame
    recording: Recording         # current recording
    network_client: Client       # network client
    height: float                # height in meters

    def __init__(self):
        self.h_fov = None
        self.v_fov = None

    def set_horizontal_fov(self, fov: float):
        """Adjust the camera's horizontal field of view """

        if self.h_fov is None:
            raise Exception(
                'This camera does not support horizontal field of view.')
        if fov < self.h_fov.min or fov > self.h_fov.max:
            raise Exception('Horizontal field of view out of bounds.')

        self.h_fov.value = fov

    def set_vertical_fov(self, fov: float):
        """Adjust the camera's vertical field of view """

        if self.v_fov is None:
            raise Exception(
                'This camera does not support vertical field of view.')
        if fov < self.v_fov.min or fov > self.v_fov.max:
            raise Exception('Vertical field of view out of bounds.')

        self.h_fov.value = fov

    def set_frame(self, recording: Recording, frame: Frame):
        """Places the camera at a specific frame """
        self.frame = frame
        self.recording = recording

        if recording.setup is not None:
            self.set_camera_height(recording.setup.camera_height)

    def set_yaw(self, yaw: float):
        """Changes the yaw of the camera """
        self.yaw = yaw

    def set_pitch(self, pitch: float):
        """Changes the pitch of the camera """
        self.pitch = pitch

    def set_camera_height(self, height: float):
        """Changes the height (meters) of the camera with respect to the altitude 
            of the current frame.
        """
        self.height = height

    def set_network_client(self, client: Client):
        """Add networking capabilities to the camera """
        self.network_client = client

####################################################################
#               SPHERICAL CAMERA AND IMAGERY                       #
####################################################################


class SphericalImage():
    """ Spherical Image

        The result of picture taken through a spherical lens, with the 
        required information for localization.

    """
    nw_client: Client
    image_provider: ImageProvider
    image_request: ImageRequest
    frame: Frame
    recording: Recording
    camera_height: float
    result = None
    camera_model: CameraModel
    computation_provider: ComputationProvider
    geo_referenced_pixels: list
    view_parameterized_pixels: list

    def __init__(self):
        self.camera_model = None
        self.geo_referenced_pixels = {}
        self.view_parameterized_pixels = {}
        pass

    def get_geo_location_camera(self) -> GeographicLocation:
        """ Returns the geographical position of where the image was taken."""
        return GeographicLocation.from_tuple(self.frame.get_location())

    def get_geo_location(self) -> GeographicLocation:
        """ Returns the geographical position on the ground plane."""
        geo = GeographicLocation.from_tuple(self.frame.get_location())
        geo.alt -= self.get_camera_height()

    def get_camera_height(self):
        """ Returns the camera height with respect to the altitude of the geographical altitude."""
        return self.camera_height

    def get_resolution(self) -> Size:
        """ Returns the image resolution (width,height)"""
        # appareturn Size(self.result.w,self.result.h)
        return self.image_request.size

    def get_field_of_view(self) -> float:
        """ Returns the image field of view (horizontal,vertical)"""
        return float(self.image_request.fov)

    def set_network_configuration(self, nw_client, provider, computation, request, camera_height, frame, recording):
        """ Initializes the SphericalImage for network usage.

            Additional computations are available when a SphericalImage has been captured
            using a SphericalCamera with a connection to the media-server. In almost all cases
            the SphericalCamera's network configuration is automatically configured by the SphericalCamera.                            
        """
        self.nw_client = nw_client
        self.image_provider = provider
        self.computation_provider = computation
        self.image_request = request
        self.camera_height = camera_height
        self.frame = frame
        self.recording = recording

    def fetch(self):
        """ Fetch the internal contents of the image.

            The function fetch allows the user to choose when to fetch the actual imagery data
            from the media-server. By default the SphericalCamera::get_image(..,manual_fetch:bool) function will perform 
            the fetch automatically, however changing the manual_fetch:bool variable it is possible to control when to 
            fetch the actual data.
        """

        client_result = self.nw_client.fetch(self.image_request)
        self.result = self.image_provider.fetch(client_result)

    def get_image(self) -> io.BytesIO:
        return self.result.image

    def get_camera_model(self) -> CameraModel:
        """ Returns/Creates a camera model from the current position.

            The CameraModel(EnuModel) can be used to make inferences as well as generating 
            points and polygons in carthesian space.
        """
        if self.camera_model == None:
            camera_location = self.frame.get_location()
            camera_heading = self.frame.heading
            if self.recording.setup is not None:
                self.camera_model = CameraModel.with_leverarms(camera_location, camera_heading,
                                                               self.recording.setup.lever_arm)
            else:
                self.camera_model = CameraModel(
                    camera_location, camera_heading)

        return self.camera_model

    def get_view_parameterized_pixel(self, pixel: Pixel) -> ViewParameterizedPixel:
        """ Returns a ViewParameterizedPixel

            This function transfroms coordinates x,y (col,row)from pixel space to a
            postition and view direction (ViewParameterizedPixel).
            A series of labelled GeoReferencedPixels/ViewParameterizedPixels can be used 
            for triangulation to obtain a GeographicLocation with higher precision.      
        """
        vpp = ViewParameterizedPixel()
        vpp.name = None
        vpp.pixel_location = pixel
        vpp.frame_index = self.frame.index
        vpp.recording_id = self.recording.id

        request_builder = ComputationRequestBuilder(
            self.frame.recordingid, self.frame.uuid)
        request = self.nw_client.fetch(request_builder.build(self.get_resolution(
        ), self.image_request.direction, self.get_field_of_view(), pixel.col, pixel.row))
        result = self.computation_provider.fetch(request)

        # yaw is with respect to north
        vpp.viewing_parameters = PositionVector(*result.values())

        return vpp

    def project_pixel_on_ground_surface(self, pixel: Pixel) -> GeoReferencedPixel:
        """ Returns a GeoReferencedPixel

            This function maps coordinates x,y (col,row)from pixel space into Geographical
            space, with the corresponding angles and postion with respect to the camera.
            It does so by assuming that the horizontal plane is uniform over the entire distance,
            and that the pixel of interest can be mapped on the surface. 
            This function is typically used to get an estimate of an object in the image that rests 
            on the ground floor. A series of labelled GeoReferencedPixels/ViewParameterizedPixels can be used 
            for triangulation to obtain a GeographicLocation with higher precision.  
        """

        grp = GeoReferencedPixel(self.get_view_parameterized_pixel(pixel))

        if not -90.0 < grp.viewing_parameters.pitch < 0.0:
            raise Exception(
                'Pitch {p:2f} is out of range [-90.0, 0.0]'.format(p=grp.viewing_parameters.pitch))

        angle = 90 + grp.viewing_parameters.pitch

        ground_north_vector = math.tan(
            math.radians(angle)) * self.get_camera_height()
        vector = SphericalImage.rotate(
            0, ground_north_vector, 0, 0, math.radians(- grp.viewing_parameters.yaw))
        camera_model = self.get_camera_model()
        grp.distance = ground_north_vector
        grp.relative_loc = RelativeLocation(*vector, -self.get_camera_height())
        grp.geo_location = GeographicLocation(
            *camera_model.to_geodetic(grp.relative_loc.location()))

        return grp

    def store_geo_referenced_pixel(self, grp: GeoReferencedPixel):
        """ Store the GeoReferencedPixel"""
        self.geo_referenced_pixels[grp.name] = grp

    def store_view_parameterized_pixel(self, vvp: ViewParameterizedPixel):
        """ Store the ViewParameterizedPixel"""
        self.view_parameterized_pixels[vvp.name] = vvp

    def rotate(x, y, xo, yo, theta):  # rotate x,y around xo,yo by theta (rad)
        xr = math.cos(theta)*(x-xo)-math.sin(theta)*(y-yo) + xo
        yr = math.sin(theta)*(x-xo)+math.cos(theta)*(y-yo) + yo
        return [xr, yr]

    @staticmethod
    def triangulate(label, list_of_images) -> GeographicLocation:
        """ Triangulate using the stored labelled 
            ViewParameterizedPixels and GeoReferencedPixels.
        """
        viewing_params_with_label = []

        for img in list_of_images:
            for name in img.geo_referenced_pixels:
                grp: GeoReferencedPixel = img.geo_referenced_pixels[name]
                if grp.name == label:
                    viewing_params_with_label.append(grp.viewing_parameters)

            for name in img.view_parameterized_pixels:
                vvp: ViewParameterizedPixel = img.view_parameterized_pixels[name]
                if vvp.name == label:
                    viewing_params_with_label.append(vvp.viewing_parameters)

        loc = Geographic.triangulate(viewing_params_with_label)
        return GeographicLocation(loc[1], loc[0], loc[2])


class SphericalCamera(Camera):
    """Camera with a spherical lens.

    This class describes a camera with a spherical lens based on the gnomonic projection 
    https://en.wikipedia.org/wiki/Gnomonic_projection. 

    The lens has the following properties:
        horizontal field of view [0,90]  <--- @todo marten
        vertical field of view [0,90]    <--- @todo marten

    """
    yaw: float
    pitch: float

    def __init__(self):
        """Construct a Spherical camera. """
        super().__init__()
        self.h_fov = FieldOfView(90.0, 1.0, 160.0)
        self.set_pitch(0)
        self.set_yaw(0)

    def set_pitch(self, pitch: float):
        """Adjust the camera's pitch

            The pitch range: 
            looking down:   -90 deg
            looking up  :    90 deg

        """
        self.pitch = pitch

    def set_yaw(self, yaw: float):
        """Adjust the camera's yaw

            The yaw range: 
            max left :  -180 deg
            max right:   180 deg

        """
        self.yaw = yaw

    def set_horizontal_fov(self, fov: float):
        """Adjust the camera's horizontal field of view """
        super().set_horizontal_fov(fov)

    def set_frame(self, recording: Recording, frame: Frame):
        """Places the camera at a specific frame """
        super().set_frame(recording, frame)

    def set_camera_height(self, height: float):
        """Changes the height (meters) of the camera with respect to the altitude 
            of the current frame.
        """
        super().set_camera_height(height)

    def set_network_client(self, client: Client):
        """Add networking capabilities to the camera """
        super().set_network_client(client)

    def look_at(self, geo_location: GeographicLocation):
        """Set the view direction yaw/pitch of the current frame to a geographic location"""
        if self.recording.setup is not None:
            camera_model = CameraModel.with_leverarms(self.frame.get_location(), self.frame.heading,
                                                      self.recording.setup.lever_arm)
        else:
            camera_model = CameraModel(
                self.frame.get_location(), self.frame.heading)

        yaw, pitch = camera_model.look_at(
            [geo_location.lon, geo_location.lat, geo_location.alt])
        self.set_yaw(yaw - self.frame.heading)
        self.set_pitch(pitch)

    def look_at_all(self, geo_locations: [GeographicLocation], width: int):

        if not horus_geometries_found:
            raise Exception('Function not supported, requires horus_geometries.')

        if self.recording.setup is not None:
            camera_model = CameraModel.with_leverarms(self.frame.get_location(), self.frame.heading,
                                                      self.recording.setup.lever_arm)
        else:
            camera_model = CameraModel(
                self.frame.get_location(), self.frame.heading)

        gp = Geometry_proj()
        proj = gp.Projection(camera_model, False)

        geom = gp.to_enu(gp.geographic_to_polygon(geo_locations), proj).geometry
        c = geom.centroid
        bearing = math.degrees(math.atan2(c.x, c.y))

        geom = gp.rotate(geom, proj, math.degrees(bearing))
        bb = gp.bounding_box(geom)
        br, tl = [gp.to_point(p) for p in bb.exterior.coords[:-1:2]]
        z = geom.exterior.coords[0][2]
        c = bb.centroid
        ct = gp.to_point([c.x, tl.y, z])
        cb = gp.to_point([c.x, br.y, z])

        vfov = camera_model.angle_between(*ct.coords, *cb.coords)
        hfov = camera_model.angle_between([0, cb.y, cb.z], [max(abs(tl.x), br.x), cb.y, z]) * 2
        pitch = camera_model.angle_between(*ct.coords, [0, 0, 1]) + vfov / 2
        height = int(math.tan(math.radians(vfov) / 2) * width / math.tan(math.radians(hfov) / 2))

        self.set_yaw(360 - self.frame.heading + bearing)
        self.set_pitch(90 - pitch)
        self.set_horizontal_fov(hfov)

        return Size(width, height)

    def crop_to_geometry(self, image:SphericalImage, geo_locations: [GeographicLocation], draw_geometry: bool = False) -> SphericalImage:
        
        if not horus_geometries_found:
            raise Exception('Function not supported, requires horus_geometries.')

        if self.recording.setup is not None:
            camera_model = CameraModel.with_leverarms(self.frame.get_location(), self.frame.heading,
                                            self.recording.setup.lever_arm)
        else:
            camera_model = CameraModel(
                self.frame.get_location(), self.frame.heading)

        size = image.image_request.size
        computation_request_builder = ComputationRequestBuilder('project')
        
        points = []
        for p in [*geo_locations, geo_locations[0]]:
            y, p = camera_model.look_at([p.lon, p.lat, p.alt])
            computation_request = image.nw_client.fetch(computation_request_builder.build(size, Direction(y - self.frame.heading, p), 
                                        self.h_fov.value, direction0 = Direction(self.yaw, self.pitch)))
            result = image.computation_provider.fetch(computation_request)
            if 'error' in result:
                print(result['error'])
            else:
                x, y = result.values()
                points.append((x, y))

        if len(points) > 0:
            with Image.open(image.get_image()) as im:
                if draw_geometry:
                    draw = ImageDraw.Draw(im)
                    for i in range(len(points) - 1):
                        draw.line((points[i], points[i + 1]), fill=(12, 255, 36))

                out = BytesIO()
                im.crop(Geometry_proj().to_polygon(points).bounds).save(out, format="JPEG", quality=95)
                image.result = ImageProvider.Result(out, [0, 0, size.width, size.height], size.width, size.height)

        return image
    
    def acquire(self, size: Size, manual_fetch: bool = False) -> SphericalImage:
        """Acquire a Spherical image from the current position/configuration of the camera"""

        ver_fov = 2 * math.degrees(math.atan(size.height * math.tan(
            (math.radians(self.h_fov.value)) / 2) / size.width))

        if not self.h_fov.min < ver_fov < self.h_fov.max:
            raise Exception('Vertical fov {ver_fov:2f} is out of range [{min:1f}, [{max:1f}]]'.format(
                ver_fov=ver_fov, min=self.h_fov.min, max=self.h_fov.max))

        if self.network_client != None:
            request_builder = ImageRequestBuilder(
                self.frame.recordingid, self.frame.uuid)
            request = request_builder.build_spherical(
                size, Direction(self.yaw, self.pitch), self.h_fov.value)

            image = SphericalImage()
            image.set_network_configuration(self.network_client, ImageProvider(
            ), ComputationProvider(), request, self.height, self.frame, self.recording)

            if not manual_fetch:
                image.fetch()

            return image

        return None



"""Horus database"""
# Copyright(C) 2022 Horus View and Explore B.V.

import os
import pathlib
import glob
from xml.etree import ElementTree
from horus_db import Recording, Recordings, Frame, Frames

from typing import NamedTuple

# Dep on the sqlite/spatialite for geosuiteDb
try:
    import spatialite
    from shapely import wkt, wkb
except ModuleNotFoundError as e:
    print(f"Install module '{e.name}' to use module 'horus_spatialite'.")
    raise e


class Spatialite:
    """
    Database used for making annotations using the Horus Geo Suite
    """
    class Field_info(NamedTuple):
        idx: int
        name: str
        type: str

    recordings = {}
    blob_containing_geometry = {}
    recording_field_name: str = None
    frame_index_field_name: str = None
    geometry_field_name = "the_geom"
    table_name: str
    field_info_map: {str, Field_info} = None

    FIELD_INFO_IDX = 0
    FIELD_INFO_NAME = 1
    FIELD_INFO_TYPE = 2
    FIELD_NAME_GEOM = "GEOMETRY"
    FIELD_NAME_BLOB = "BLOB"

    # --- Remote Database
    RD_connection = None
    spatialite_RD_recording_map = {}

    # ------ Recordings On Disk  ------
    ROD_frame_location_guids = []
    ROD_recordings_root_folder: str = None
    spatialite_ROD_recording_map = {}

    def __init__(self, filename):
        self.filename = filename
        self.table_name = self.__table_name_from_file__()

    def set_recordings_on_disk_root_folder(self, path):
        self.ROD_recordings_root_folder = path

    def set_recording_field(self, field_name):
        self.recording_field_name = field_name

    def set_frame_index_field(self, field_name):
        self.frame_index_field_name = field_name

    def set_remote_db_connection(self, connection):
        self.RD_connection = connection

    def open(self):
        self.conn = spatialite.connect(self.filename)

    def close(self):
        self.conn.close()

    def get_cursor(self):
        return self.conn.cursor()

    def check_cursor(self, cursor):
        if cursor is None:
            cursor = self.get_cursor()
            return True, cursor
        return False, cursor

    def version(self):
        return self.conn.execute('SELECT spatialite_version()').fetchone()[0]

    def get_table_name(self):
        return self.table_name

    def set_table_name(self, name):
        self.table_name = name

    def get_field_query(self, cursor):
        cursor.execute("SELECT sql FROM sqlite_master WHERE tbl_name = '" +
                       self.table_name + "' AND type = 'table'")

    def get_table_info_query(self, cursor):
        cursor.execute("PRAGMA table_info(\""+self.table_name+"\")")

    def get_table_names(self, cursor):
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")

    def get_recordings(self, cursor):
        cursor.execute("SELECT DISTINCT	" + self.recording_field_name +
                       " FROM \"" + self.table_name + "\";")
    
    def set_geometry_field_name(self, field_name):
        self.geometry_field_name = field_name;

    def blob_contains_geometry(self,field_name):
        self.blob_containing_geometry[field_name] = True

    def get_field_names_map(self, cursor=None) -> {str: Field_info}:
        if not self.field_info_map is None:
            return self.field_info_map

        cleanup, cursor = self.check_cursor(cursor)
        self.get_table_info_query(cursor)
        records = cursor.fetchall()

        data = {row[self.FIELD_INFO_NAME]: Spatialite.Field_info(row[self.FIELD_INFO_IDX],
                                                                 row[self.FIELD_INFO_NAME],
                                                                 row[self.FIELD_INFO_TYPE]) for row in records}
        row_idx = len(data)
        data["rowid"] = Spatialite.Field_info(row_idx, "rowid", "INTEGER")

        if cleanup:
            cursor.close()

        return data

    def query(self, cursor=None, field_names_map=None, order_by_list=[]):
        cleanup, cursor = self.check_cursor(cursor)

        if field_names_map is None:
            if self.field_info_map is None:
                field_info_map = self.get_field_names_map(cursor)
            else:
                field_names_map = self.field_info_map

        fields = []
        for k, field in field_names_map.items():
            if field.type == self.FIELD_NAME_GEOM:
                fields.append("AsText(" + k + ")")
            else:
                fields.append(k)

        myselect = ','.join(fields)

        query_ = "SELECT " + myselect + " FROM \""+self.table_name+"\""

        add = " "
        for x in order_by_list:
            query_ += add + "ORDER BY " + x + " ASC"
            add = ","

        query_ += ";"
        cursor.execute(query_)
        return cursor

    def get_geometry(self, cursor, field_names_map=None):
        if field_names_map == None:
            field_names_map = self.field_info_map
        geoms = {}
        for k, field in field_names_map.items():
            if field.type == self.FIELD_NAME_GEOM:
                if cursor[field.idx] != None:
                    geoms[k] = wkt.loads(cursor[field.idx])
            elif field.type == self.FIELD_NAME_BLOB:                
                if field.name in self.blob_containing_geometry:                    
                    if cursor[field.idx] != None:                        
                        geoms[k] = wkb.loads(cursor[field.idx])                            
        return geoms

    def get_matched_frames_iterator(self):
        cursor = self.get_cursor()
        orderby = []
        if self.recording_field_name in self.field_info_map:
            orderby.append(self.recording_field_name)
        self.query(cursor, self.field_info_map, orderby)
        return FrameMatchedIterator(cursor, self.RD_connection, self)

    def resolve(self):
        # obtain the fields with the table appended with rowid
        cursor = self.get_cursor()
        self.field_info_map = self.get_field_names_map(cursor)

        if self.recording_field_name != None and self.recording_field_name in self.field_info_map:
            # Get all the unique recordings in spatialite
            self.get_recordings(cursor)
            data = [x[0] for x in cursor.fetchall() if x[0] != None]
            for path in data:
                __FIX_ME__ = path.replace("\\", "/")  # windows unix hell
                p = pathlib.PurePath(__FIX_ME__)
                if p.suffix.strip() == ".idx":
                    p = p.parent
                self.recordings[path] = p.name

        # Resolve relation local recording(field) => remote
        if self.recording_field_name != None and self.RD_connection != None:
            self.spatialite_RD_recording_map = self.resolve_remote_recording(
                self.recording_field_name, self.RD_connection, cursor)

        # Resolve relation local recording(field => disk)
        if self.ROD_recordings_root_folder != None:
            self.resolve_on_disk_recording(self.ROD_recordings_root_folder)

    def match_paths(self, r, p, candidates):
        r1 = r.split('\\')
        c1 = [c.split('/') for c in candidates]

        while r1[-1] != p:
            r1.pop()

        try:
            while True:
                r1.pop()
                matched = 0
                index = 0
                idx = 0
                for c in c1:
                    c.pop()
                    if c[-1] == r1[-1]:
                        matched += 1
                        index = idx
                    idx += 1

                if matched == 1:
                    return candidates[index]
        except:
            return None

    def resolve_on_disk_recording(self, recording_folder: str):
        for k, v in self.recordings.items():
            folder = recording_folder+"**/"+v

            candidates = glob.glob(folder, recursive=True)
            matched = self.match_paths(k, v, candidates)
            if matched != None:
                self.spatialite_ROD_recording_map[k] = matched

    def resolve_remote_recording(self, recording_field_name: str, connection, cursor):
        cleanup, cursor = self.check_cursor(cursor)
        resolved = {}

        for k, v in self.recordings.items():
            recordings = Recordings(connection)
            cursor2 = Recording.query(
                recordings, directory_like=v, order_by="id")

            rec = next(cursor2, None)
            recordings.get_setup(rec)

            recs = []
            while (rec) != None:
                recs.append(rec)
                rec = next(cursor2, None)

            while len(recs) != 1:
                raise Exception(
                    "Spatialite::resolve_recording Implement duplicate!")

            resolved[k] = recs[0]

        if cleanup:
            cursor.close()

        return resolved

    def resolve_frames(self, recording_name):
        for k, v in self.spatialite_ROD_recording_map.items():
            if k == recording_name:
                tree = ElementTree.parse(v+"/frames.xml")
                root = tree.getroot()
                a = tree.findall(
                    "./{http://tempuri.org/FramesDataSet.xsd}Location/{http://tempuri.org/FramesDataSet.xsd}GUID")
                frame_location_guids = [x.text for x in a]
                return frame_location_guids

    def __table_name_from_file__(self):
        return os.path.splitext(os.path.basename(self.filename))[0]

    def show_info(self):

        print("\n------- Spatialite --------")
        print("Version: ", self.version())
        print("Table: ", self.get_table_name())
        print("\n")

        print("\n------- Field Info --------")
        for k, field_info in self.field_info_map.items():
            print(field_info.idx, " ", field_info.name, "  ", field_info.type)
        print("\n")

        cursor = self.get_cursor()

        print("\n------- Browse Data --------")
        self.query(cursor, self.field_info_map)

        entries = 4
        for row in cursor:
            if entries > 0:
                entries -= 1
                geoms = self.get_geometry(row, self.field_info_map)                
                print("--------------------")
                for k, field_info in self.field_info_map.items():
                    if k in geoms:
                        print(field_info.idx, ":", geoms[k]) 
                    else:
                        print(field_info.idx, ":", row[field_info.idx])

        print("\n-------  Recordings / Spatialite --------")
        for k, v in self.recordings.items():
            print(k, " -> ", v)

        print("\n-------  Recording / Remote relation --------")
        for k, rec in self.spatialite_RD_recording_map.items():
            print(k, " -> ", rec)

        print("\n-------  Recording / On Disk relation --------")
        for k, v in self.spatialite_ROD_recording_map.items():
            print(k, " -> ", v)


class FrameMatchedIterator:

    class MatchedFrame:
        spatialite_cursor = None
        frame: Frame = None
        recording: Recording = None
        properties = {}
        metadata = {}

        def dump(self):
            print("spatialite_cursor", self.spatialite_cursor)
            print("Frame", self.frame)
            print("Recording", self.recording)
            print("Properties", self.properties)

        def oke(self):
            return self.spatialite_cursor != None and self.frame != None and self.recording != None

    spatialite_db: Spatialite
    current_recording: str = None
    static_recording: Recording = None
    recordings_list: [str, Recording] = {}

    guids: [str] = None

    use_recording_field = False
    use_frame_index_field = False
    use_static_recording = False

    recording_field_idx: int
    frame_index_field_idx: int

    def __init__(self, cursor, connection, spatialite_db: Spatialite):
        self.cursor = cursor
        self.connection = connection
        self.spatialite_db = spatialite_db
        self.frames = Frames(connection)
        self.recordings = Recordings(connection)

        # Can we use the Recording field
        if spatialite_db.recording_field_name != None:
            if self.spatialite_db.recording_field_name in self.spatialite_db.field_info_map:
                self.use_recording_field = True
                self.recording_field_idx = self.spatialite_db.field_info_map[
                    self.spatialite_db.recording_field_name].idx

        # Can we use Frame index field
        if spatialite_db.frame_index_field_name != None:
            if self.spatialite_db.frame_index_field_name in self.spatialite_db.field_info_map:
                self.use_frame_index_field = True
                self.frame_index_field_idx = self.spatialite_db.field_info_map[
                    self.spatialite_db.frame_index_field_name].idx

    def set_static_recording_by_id(self, id):
        recordings = Recordings(self.connection)
        cursor = Recording.query(recordings, id=id)
        recording = next(cursor, None)
        if recording != None:
            recordings.get_setup(recording)
            self.use_static_recording = True
            self.static_recording = recording

    def __iter__(self):
        return self

    def __next__(self):
        frame = next(self.cursor, None)

        while frame != None:

            # Build up the matched frame
            f = self.MatchedFrame()

            use_recording_field = self.use_recording_field
            use_frame_index_field = self.use_frame_index_field
            frame_guid = None
            frame_index = None

            if use_recording_field:
                if frame[self.recording_field_idx] == None:
                    use_recording_field = False
                else:
                    if frame[self.recording_field_idx] != self.current_recording:
                        self.current_recording = frame[self.recording_field_idx]
                        self.guids = self.spatialite_db.resolve_frames(
                            self.current_recording)

            if use_frame_index_field:
                if frame[self.frame_index_field_idx] == None:
                    use_frame_index_field = False
                else:
                    frame_index = int(frame[self.frame_index_field_idx])
                    # Try to be more precise with GUID
                    if self.guids != None and len(self.guids) >= frame_index:
                        frame_guid = self.guids[int(
                            frame[self.frame_index_field_idx])]

            if frame_guid != None:
                f.properties["guid"] = frame_guid

            if frame_index != None:
                f.properties["index"] = frame_index

            # Allow static recordings set_static_recording_by_id(..)
            if self.use_static_recording:
                f.properties["recordingid"] = self.static_recording.id

            elif not self.current_recording is None:
                if not self.spatialite_db.spatialite_RD_recording_map is None:
                    if self.current_recording in self.spatialite_db.spatialite_RD_recording_map:
                        f.properties["recordingid"] = self.spatialite_db.spatialite_RD_recording_map[self.current_recording].id

            has_guid = "guid" in f.properties
            has_frame_index = "index" in f.properties
            has_recording_id = "recordingid" in f.properties

            if (has_recording_id and has_frame_index) or has_guid:
                cursor = self.frames.query(**f.properties)
            else:
                geom = self.spatialite_db.get_geometry(
                    frame)[self.spatialite_db.geometry_field_name]

                distance_min = 5
                distance_max = 10
                cursor = self.frames.query(within=(*geom.centroid.coords, distance_max),
                                           **f.properties,
                                           distance=(*geom.centroid.coords, "> %s",
                                                     distance_min), limit=1)

            f.spatialite_cursor = frame
            f.frame = Frame(cursor)

            if (f.frame == None):
                return f

            if not f.frame.recordingid in self.recordings_list:
                temp_rec = next(Recording.query(
                    self.recordings, id=f.frame.recordingid))
                self.recordings.get_setup(temp_rec)
                self.recordings_list[f.frame.recordingid] = temp_rec

            f.recording = self.recordings_list[f.frame.recordingid]

            self.add_metadata_from_remote_db(f, {"file", "path"})

            return f
        else:
            raise StopIteration

    def add_metadata_from_remote_db(self, f, keys):
        iie_query = "select * from frames,image_index_entries where image_index_entries.frame_id = " + \
            str(f.frame.id) + " and frames.recordingid = " + \
            str(f.recording.id)

        __cursor__ = self.connection.cursor()
        __cursor__.execute(iie_query)

        values = __cursor__.fetchone()
        field_names = [field[0] for field in __cursor__.description]
        row = dict(zip(field_names, values))

        for k in keys:
            if k in row:
                f.metadata["image_index_entries."+k] = row[k]



from horus_gis import SchemaProvider, PositionVector, Geographic
from horus_geopandas import HorusGeoDataFrame
import geopandas as gpd
import os

### Read Data ####
sp = SchemaProvider()
schema = sp.merge(sp.single_measurement(),sp.clustering())
database = HorusGeoDataFrame(schema)

single_measurement_path = "output/single_measurement.shp"
if not os.path.exists(single_measurement_path):
    print(f"File '{single_measurement_path}' not found.\nRun 'horus_media_examples.spherical_camera_single_measurement' first.")
    exit(1)
database.append_file("output/single_measurement_clusters.shp")

dataframe = database.dataframe

### Reconstruct PositionVector
surf_vpp_per_cluster =  {k: [] for k in dataframe['clstr_id'].unique()}
geolocation_per_cluster = {}
confidence_per_cluster ={}


for index, row in dataframe.iterrows(): # Looping over all points
    ### Surface info =  ViewParameterizedPixel
    pos_vector = PositionVector(
        dataframe.at[index, "cam_lat"],
        dataframe.at[index, "cam_lon"],
        dataframe.at[index, "cam_alt"],
        dataframe.at[index, "surf_yaw"],
        dataframe.at[index, "surf_pitch"])
    id =  dataframe.at[index, "clstr_id"]
    surf_vpp_per_cluster[id].append(pos_vector)


### Triangulate
for vpp in surf_vpp_per_cluster:
    pos_vectors = surf_vpp_per_cluster[vpp]
    geolocation_per_cluster[vpp] = Geographic.triangulate(pos_vectors)
    confidence_per_cluster[vpp] = 0.7 # some metric

### Write Data
schema = sp.merge(sp.geometry_3dpoint(), sp.clustering())
database = HorusGeoDataFrame(schema)

for vpp in surf_vpp_per_cluster:
    lat = geolocation_per_cluster[vpp][0]
    lon = geolocation_per_cluster[vpp][1]
    alt = geolocation_per_cluster[vpp][2]

    record = database.new_frame()
    record["geometry"] = gpd.points_from_xy(x=[lon], y=[lat], z=[alt])
    record["clstr_id"] = vpp
    record["clstr_conf"] = confidence_per_cluster[vpp]
    database.add_frame(record)




database.write_shapefile("output/cluster_locations.shp")


from horus_camera import SphericalCamera
from horus_gis import GeographicLocation
from horus_db import Recordings, Frames, Frame
from horus_spatialite import Spatialite, FrameMatchedIterator
from pyproj import Geod
from horus_geometries import Geometry_proj
import geopandas as gpd
import pandas as pd
import sys
from os import path
import traceback

from . import util

parser = util.create_argument_parser()

parser.add_argument("--sqlite-db", type=str,
                    help="the geosuite_database name")

parser.add_argument("--sqlite-framenr", type=str,
                    help="the field specifying the framenr")

parser.add_argument("--sqlite-recording", type=str,
                    help="the field specifying the recordingname")

parser.add_argument("--sqlite-geometry", type=str,
                    help="the field specifying an alternative geometry blob")

# Not tested
parser.add_argument("--recordings-on-disk", type=str,
                    help="Optionally provide a recording folder")

# Not tested
parser.add_argument("--recording-id", type=int,
                    help="Optionally provide a the static recording id.")


util.add_database_arguments(parser)
util.add_server_arguments(parser)


args = parser.parse_args()
client = util.get_client(args)
connection = util.get_connection(args)
recordings = Recordings(connection)

if args.sqlite_db is None:
    print("A sqlite database should be provided")
    exit()

# sqlite_frame_idx_field = "Frame_numb"
output_database = None
geod = Geod(ellps="WGS84")


class Look_at:
    """
        Simple class that describes what geometry the virtual camera
        should look at and from which frame.        
    """
    frame: Frame = None
    geometry = None

    def __init__(self, frame, geometry):
        self.frame = frame
        self.geometry = geometry
        self.gp = Geometry_proj()

    def to_geographic_loc_list(self):
        return self.gp.to_geographic(self.geometry) 


def add_to_output(frame, geometry, filename, nr, matched_frame):
    """
       Add the curreont geometry and attributes to the output database
    """
    global output_database

    record = {}
    record["geometry"] = gpd.GeoSeries(geometry)
    record["snapshot"] = filename
    record["sub_id"] = nr
    record["frame_index"] = frame.index
    record["recording_id"] = frame.recordingid

    for k, v in matched_frame.metadata.items():
        record[k] = v

    gdf = gpd.GeoDataFrame(record)

    if output_database is None:
        output_database = gdf
    else:
        output_database = pd.concat([output_database, gdf])


def try_find_frame_within(geometry, mf: FrameMatchedIterator.MatchedFrame, distance_min, distance_max):
    """
       Try to find a frame from the same recording that is within (distance_min,distance_max)
    """
    frames = Frames(connection)
    cursor = frames.query(within=(*geometry.centroid.coords, distance_max),
                          recordingid=mf.recording.id,
                          distance=(*geometry.centroid.coords, "> %s",
                                    distance_min), limit=1)

    if (cursor != None):
        return Frame(cursor)
    return None


def look_at_point(point, mf: FrameMatchedIterator.MatchedFrame, side):
    """
        Transforms a 'Point' geometry into a square of sides 'side'x'side'
        in meters
    """
    gp = Geometry_proj()
    square = gp.point_to_square(point, 0.5)
    return [Look_at(mf.frame, square)]


def look_at_linestring(linestring, mf: FrameMatchedIterator.MatchedFrame, max_length, distance_min, distance_max, offset):
    """
        Transforms a 'LineString' geometry into sub LineStrings of maximum length
        of 'max_length' in meters.
    """
    gp = Geometry_proj()
    geom_length = geod.geometry_length(linestring)
    
    if geod.geometry_length(linestring) > max_length:
        
        linestrings = gp.split_linestring(linestring,max_length)
                
        look_at_all_sub_string = []
        for w in linestrings:

            frame = try_find_frame_within(w, mf, distance_min, distance_max)

            if not frame is None:
                mf.frame = frame

            look_at_all_sub_string.append(look_at_linestring(
                w, mf, max_length, distance_min, distance_max, offset)[0])

        return look_at_all_sub_string

    return [Look_at(mf.frame,gp.buffer(linestring, offset))]    


def look_at_polygon(polygon, mf: FrameMatchedIterator.MatchedFrame):
    """
        Transforms a 'Polygon' geometry into a collection of Polygons..
    """
    return [Look_at(mf.frame, polygon)]


def take_shaphost(db, mf: FrameMatchedIterator.MatchedFrame, geod):

    geometries = []
    geo = db.get_geometry(mf.spatialite_cursor)[db.geometry_field_name]
    
    if geo.type.startswith('Multi'):
        for g in geo.geoms:
            geometries.append(g)
    else:
        geometries.append(geo)

    width = 800
    look_at_all: [Look_at] = []

    for geom in geometries:
        if geom.type == "Polygon":
            look_at_all = [*look_at_all, *look_at_polygon(geom, mf)]
        elif geom.type == "LineString":
            look_at_all = [*look_at_all, *look_at_linestring(geom, mf, 5, 5, 20, 0.1)]       
        elif geom.type == "Point":
            look_at_all = [*look_at_all, *look_at_point(geom,mf, 0.3) ]       
        else:
            print("Not supported", geom.type)

    nr_shaphosts = len(look_at_all)

    for x, look_at in enumerate(look_at_all):
        look_at_geometry = look_at.to_geographic_loc_list()

        sp_camera.set_frame(mf.recording, look_at.frame)
        size = sp_camera.look_at_all(look_at_geometry, width)
        spherical_image = sp_camera.crop_to_geometry(sp_camera.acquire(size), look_at_geometry)

        # Write output
        db_id = mf.spatialite_cursor[db.field_info_map["rowid"].idx]
        print("Snapshot:", "db id", db_id, "nr",
              x+1, "/", nr_shaphosts, geom.type)

        filename = 'output/snapshot_db_id:' + \
            str(db_id) + '_' + str(x+1) + '_' + geom.type + '.jpeg'

        add_to_output(look_at.frame, look_at.geometry, filename, x+1, mf)

        with open(filename, 'wb') as image_file:
            image_file.write(spherical_image.get_image().getvalue())
            spherical_image.get_image().close()


db = Spatialite(args.sqlite_db)


if not args.recordings_on_disk is None:
    db.set_recordings_on_disk_root_folder(args.recordings_on_disk)

if not args.sqlite_recording is None:
    db.set_recording_field(args.sqlite_recording)

if not args.sqlite_framenr is None:
    db.set_frame_index_field(args.sqlite_framenr)

if not args.sqlite_geometry is None:
    db.set_geometry_field_name(args.sqlite_geometry)
    db.blob_contains_geometry(args.sqlite_geometry)

db.set_remote_db_connection(connection)

db.open()
db.resolve()
db.show_info()

sp_camera = SphericalCamera()
sp_camera.set_network_client(client)

fmi: FrameMatchedIterator = db.get_matched_frames_iterator()

if (not args.recording_id is None):
    fmi.set_static_recording_by_id(args.recording_id)


mf: FrameMatchedIterator.MatchedFrame = next(fmi, None)

while mf != None:

    try:
        if mf.oke():
            take_shaphost(db, mf, geod)
        else:
            print("\nIncomplete match")
            print(mf.dump())
    except Exception as e:
        print("Exception", e)
        print("Properties: ", mf.properties)
        print("Metadata: ", mf.metadata)
        pass
    mf = next(fmi, None)

db.close()

if not output_database is None:
    output_database.to_file("output/snapshots.geojson", driver='GeoJSON')
